"""
Test cases for PdgParticle.
"""
from __future__ import print_function

import unittest

import pdg
from pdg.errors import PdgAmbiguousValueError


class TestData(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.api = pdg.connect(pedantic=True)

    def test_properties(self):
        self.assertRaises(PdgAmbiguousValueError, getattr, self.api.get('Q007')[0], 'mass')

    def test_summary(self):
        self.assertRaises(PdgAmbiguousValueError, self.api.get('S013D').best_summary)



"""
Test cases for PdgParticle.
"""
from __future__ import print_function

import unittest

import pdg
from pdg.decay import PdgBranchingFraction, PdgDecayProduct
from pdg.errors import PdgAmbiguousValueError, PdgNoDataError
from pdg.particle import PdgParticle


class TestData(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.api = pdg.connect(pedantic=False)

    def setUp(self):
        self.api.pedantic = False

    def test_all_particle_data(self):
        n_errors = 0
        for plist in self.api.get_particles():
            for p in plist:
                try:
                    p._get_particle_data()
                except Exception as e:
                    n_errors += 1
                    print(e)
        self.assertEqual(n_errors, 0)

    def test_name(self):
        self.assertEqual(self.api.get_particle_by_name('p').mcid, 2212)
        self.assertEqual(self.api.get_particle_by_name('pbar').mcid, -2212)

    def test_mcid(self):
        self.assertEqual(self.api.get_particle_by_mcid(5).name, 'b')
        self.assertEqual(self.api.get_particle_by_mcid(-5).name, 'bbar')
        self.assertEqual(self.api.get_particle_by_mcid(5).mcid, 5)
        self.assertEqual(self.api.get_particle_by_mcid(-5).mcid, -5)
        self.assertEqual(self.api.get_particle_by_mcid(11).name, 'e-')
        self.assertEqual(self.api.get_particle_by_mcid(-11).name, 'e+')
        # self.assertEqual(self.api.get_particle_by_mcid(39).name, 'graviton')
        self.assertEqual(self.api.get_particle_by_mcid(100211).name, 'pi(1300)+')
        self.assertEqual(self.api.get_particle_by_mcid(-30323).name, 'K^*(1680)-')
        self.assertEqual(self.api.get_particle_by_mcid(-30323).mcid, -30323)

    def test_equivalent_names(self):
        p1 = self.api.get_particle_by_name('K(S)')
        p2 = self.api.get_particle_by_name('K(S)0')
        self.assertEqual(p1.pdgid, p2.pdgid)

    def test_generic_names(self):
        self.assertRaises(PdgAmbiguousValueError,
                          lambda: self.api.get_particle_by_name('e'))
        self.assertRaises(PdgAmbiguousValueError,
                          lambda: self.api.get_particle_by_name('pi'))
        mus = list(self.api.get_particles_by_name('mu'))
        self.assertEqual(len(mus), 2)
        self.assertEqual(mus[0].mcid, 13)
        self.assertEqual(mus[1].mcid, -13)

    def test_quantum_P(self):
        self.assertEqual(self.api.get_particle_by_name('u').quantum_P, '+')
        self.assertEqual(self.api.get_particle_by_name('ubar').quantum_P, '-')
        self.assertEqual(self.api.get_particle_by_name('t').quantum_P, '+')
        self.assertEqual(self.api.get_particle_by_name('tbar').quantum_P, '-')
        self.assertEqual(self.api.get_particle_by_name('p').quantum_P, '+')
        self.assertEqual(self.api.get_particle_by_name('pbar').quantum_P, '-')
        self.assertEqual(self.api.get_particle_by_name('n').quantum_P, '+')
        self.assertEqual(self.api.get_particle_by_name('nbar').quantum_P, '-')
        self.assertEqual(self.api.get_particle_by_mcid(3122).quantum_P, '+')
        self.assertEqual(self.api.get_particle_by_mcid(-3122).quantum_P, '-')

    def test_mass_2022(self):
        if '2022' not in self.api.editions:
            return
        masses = list(self.api.get('q007/2022')[0].masses())
        self.assertEqual(masses[0].pdgid, 'Q007TP/2022')
        self.assertEqual(round(masses[0].best_summary().value,9), 172.687433378)
        self.assertEqual(masses[1].pdgid, 'Q007TP2/2022')
        self.assertEqual(round(masses[1].best_summary().value,9), 162.500284698)
        self.assertEqual(masses[2].pdgid, 'Q007TP4/2022')
        self.assertEqual(round(masses[2].best_summary().value,9), 172.463424407)
        self.assertEqual(round(self.api.get('s008/2022')[0].mass,9), 0.139570391)
        self.assertEqual(round(self.api.get('s008/2022')[0].mass_error,16), 1.820071604e-07)

    def test_best_masses(self):
        self.assertEqual(round(self.api.get_particle_by_name('p').mass, 3), 0.938)
        self.assertEqual(round(self.api.get_particle_by_name('pi+').mass, 3), 0.140)
        self.assertEqual(round(self.api.get_particle_by_name('e-').mass, 6), 0.000511)
        self.assertEqual(round(self.api.get_particle_by_name('tbar').mass, 1), 172.6)
        self.assertEqual(round(self.api.get_particle_by_name('B+').mass, 2), 5.28)
        self.assertEqual(round(self.api.get_particle_by_name('Z').mass, 1), 91.2)

    def test_flags(self):
        self.assertEqual(self.api.get('S008')[0].is_boson, False)
        self.assertEqual(self.api.get('S008')[0].is_quark, False)
        self.assertEqual(self.api.get('S008')[0].is_lepton, False)
        self.assertEqual(self.api.get('S008')[0].is_meson, True)
        self.assertEqual(self.api.get('S008')[0].is_baryon, False)
        self.assertEqual(self.api.get('q007')[0].is_quark, True)
        self.assertEqual(self.api.get('s000')[0].is_boson, True)
        self.assertEqual(self.api.get('S003')[0].is_lepton, True)
        self.assertEqual(self.api.get('S041')[0].is_meson, True)
        self.assertEqual(self.api.get('S016')[0].is_baryon, True)

    def test_properties(self):
        self.assertEqual(len(list(self.api.get('S017')[0].properties('M'))), 2)

    def test_ambiguous_defaults(self):
        self.assertEqual(round(self.api.get('Q007')[0].mass, 1), 172.6)
        self.assertEqual(self.api.get('S013D').best_summary().comment, 'Assuming CPT')

    def test_best_widths_and_lifetimes(self):
        pi0 = self.api.get_particle_by_name('pi0')
        self.assertTrue(pi0.has_lifetime_entry)
        self.assertFalse(pi0.has_width_entry)
        self.assertEqual(round(pi0.lifetime * 1e17, 2), 8.43)
        self.assertEqual(round(pi0.lifetime_error * 1e17, 3), 0.134)
        self.api.pedantic = True
        self.assertRaises(PdgNoDataError, lambda: pi0.width)
        self.assertRaises(PdgNoDataError, lambda: pi0.width_error)
        self.api.pedantic = False
        self.assertEqual(round(pi0.width * 1e9, 2), 7.81)
        self.assertEqual(round(pi0.width_error * 1e9, 3), 0.125)

        W = self.api.get_particle_by_name('W+')
        self.assertTrue(W.has_width_entry)
        self.assertFalse(W.has_lifetime_entry)
        self.assertEqual(round(W.width, 3), 2.137)
        self.assertEqual(round(W.width_error, 4), 0.0537)
        self.api.pedantic = True
        self.assertRaises(PdgNoDataError, lambda: W.lifetime)
        self.assertRaises(PdgNoDataError, lambda: W.lifetime_error)
        self.api.pedantic = False
        self.assertEqual(round(W.lifetime * 1e25, 2), 3.08)
        self.assertEqual(round(W.lifetime_error * 1e25, 3), 0.077)

    def test_Kstar_892(self):
        ps = self.api.get('M018')
        ps = sorted(ps, key=lambda p: p.name)
        self.assertEqual([p.name for p in ps],
                         ['K^*(892)+', 'K^*(892)-', 'K^*(892)0', 'Kbar^*(892)0'])
        self.assertEqual([len(list(p.masses())) for p in ps],
                         [3, 3, 2, 2])
        self.assertEqual([len(list(p.widths())) for p in ps],
                         [2, 2, 1, 1])
        self.assertEqual([len(list(p.lifetimes())) for p in ps],
                         [0, 0, 0, 0])

        self.api.pedantic = True
        self.assertRaises(PdgAmbiguousValueError, lambda: ps[0].mass)
        self.assertRaises(PdgAmbiguousValueError, lambda: ps[0].width)
        self.assertRaises(PdgNoDataError, lambda: ps[0].lifetime)
        self.assertEqual(ps[0].charge, 1)

        self.api.pedantic = False
        self.assertIsNotNone(ps[0].mass)
        self.assertIsNotNone(ps[0].width)
        self.assertIsNotNone(ps[0].lifetime)
        self.assertEqual(ps[0].charge, 1)

        for self.api.pedantic in [True, False]:
            p = self.api.get_particle_by_mcid(323)
            self.assertEqual(len(list(p.masses())), 3)
            self.assertEqual(len(list(p.widths())), 2)
            self.assertEqual(len(list(p.lifetimes())), 0)
            self.assertEqual(p.charge, 1.0)
        self.api.pedantic = False
        p = self.api.get_particle_by_mcid(323)
        self.assertEqual(round(p.mass, 3), 0.892)
        self.assertEqual(round(p.width, 4), 0.0514)
        self.assertEqual(round(p.lifetime * 1e23, 2), 1.28)
        self.api.pedantic = True
        p = self.api.get_particle_by_mcid(323)
        self.assertRaises(PdgAmbiguousValueError, lambda: p.mass)
        self.assertRaises(PdgAmbiguousValueError, lambda: p.width)
        self.assertRaises(PdgNoDataError, lambda: p.lifetime)

        for self.api.pedantic in [True, False]:
            p = self.api.get_particle_by_mcid(-323)
            self.assertEqual(len(list(p.masses())), 3)
            self.assertEqual(len(list(p.widths())), 2)
            self.assertEqual(len(list(p.lifetimes())), 0)
            self.assertEqual(p.charge, -1.0)
        self.api.pedantic = False
        p = self.api.get_particle_by_mcid(-323)
        self.assertEqual(round(p.mass, 3), 0.892)
        self.assertEqual(round(p.width, 4), 0.0514)
        self.assertEqual(round(p.lifetime * 1e23, 2), 1.28)
        self.api.pedantic = True
        p = self.api.get_particle_by_mcid(-323)
        self.assertRaises(PdgAmbiguousValueError, lambda: p.mass)
        self.assertRaises(PdgAmbiguousValueError, lambda: p.width)
        self.assertRaises(PdgNoDataError, lambda: p.lifetime)

        for self.api.pedantic in [True, False]:
            p = self.api.get_particle_by_mcid(313)
            self.assertEqual(len(list(p.masses())), 2)
            self.assertEqual(len(list(p.widths())), 1)
            self.assertEqual(len(list(p.lifetimes())), 0)
            self.assertEqual(p.charge, 0.0)
        self.api.pedantic = False
        p = self.api.get_particle_by_mcid(313)
        self.assertEqual(round(p.mass, 3), 0.896)
        self.assertEqual(round(p.width, 4), 0.0473)
        self.assertEqual(round(p.lifetime * 1e23, 2), 1.39)
        self.api.pedantic = True
        p = self.api.get_particle_by_mcid(313)
        self.assertRaises(PdgAmbiguousValueError, lambda: p.mass)
        self.assertEqual(round(p.width, 4), 0.0473)
        self.assertRaises(PdgNoDataError, lambda: p.lifetime)

    def test_decay_W2PiGamma(self):
        decay = self.api.get('S043.6')
        self.assertIsInstance(decay, PdgBranchingFraction)
        self.assertTrue(decay.is_limit)
        self.assertEqual(decay.value, 1.9e-6)
        ps = decay.decay_products
        self.assertTrue(isinstance(p, PdgDecayProduct) for p in ps)

        self.assertEqual(ps[0].multiplier, 1)
        self.assertIsNone(ps[0].subdecay)
        self.assertEqual(ps[0].item.name, 'pi+')
        self.assertEqual(ps[0].item.item_type, 'P')
        self.assertTrue(ps[0].item.has_particle)
        piplus = ps[0].item.particle
        self.assertIsInstance(piplus, PdgParticle)
        self.assertEqual(piplus.pdgid, 'S008/%s' % self.api.default_edition)

        self.assertEqual(ps[1].multiplier, 1)
        self.assertIsNone(ps[1].subdecay)
        self.assertEqual(ps[1].item.name, 'gamma')
        self.assertEqual(ps[1].item.item_type, 'P')
        self.assertTrue(ps[1].item.has_particle)
        gamma = ps[1].item.particle
        self.assertIsInstance(gamma, PdgParticle)
        self.assertEqual(gamma.pdgid, 'S000/%s' % self.api.default_edition)

    def test_decay_Z2JpsiX(self):
        decay = self.api.get('S044.23')
        self.assertIsInstance(decay, PdgBranchingFraction)
        self.assertFalse(decay.is_limit)
        self.assertEqual(decay.value, 0.00351170065655418)
        ps = decay.decay_products
        self.assertTrue(isinstance(p, PdgDecayProduct) for p in ps)

        self.assertEqual(ps[0].multiplier, 1)
        self.assertIsNone(ps[0].subdecay)
        self.assertEqual(ps[0].item.name, 'J/psi(1S)')
        self.assertEqual(ps[0].item.item_type, 'P')
        self.assertTrue(ps[0].item.has_particle)
        jpsi = ps[0].item.particle
        self.assertIsInstance(jpsi, PdgParticle)
        self.assertEqual(jpsi.pdgid, 'M070/%s' % self.api.default_edition)

        self.assertEqual(ps[1].multiplier, 1)
        self.assertIsNone(ps[1].subdecay)
        self.assertEqual(ps[1].item.name, 'X')
        self.assertEqual(ps[1].item.item_type, 'I')
        self.assertFalse(ps[1].item.has_particle)

    # This tests the use of PDGITEM_MAP to get from Z to Z0
    def test_decay_H2ZGamma(self):
        decay = self.api.get('S126.6')
        self.assertIsInstance(decay, PdgBranchingFraction)
        self.assertFalse(decay.is_limit)
        self.assertEqual(decay.value, 0.0034)
        ps = decay.decay_products
        self.assertTrue(isinstance(p, PdgDecayProduct) for p in ps)

        self.assertEqual(ps[0].multiplier, 1)
        self.assertIsNone(ps[0].subdecay)
        self.assertEqual(ps[0].item.name, 'Z')
        self.assertEqual(ps[0].item.item_type, 'G')
        self.assertTrue(ps[0].item.has_particle)
        Z = ps[0].item.particle
        self.assertIsInstance(Z, PdgParticle)
        self.assertEqual(Z.pdgid, 'S044/%s' % self.api.default_edition)

        self.assertEqual(ps[1].multiplier, 1)
        self.assertIsNone(ps[1].subdecay)
        self.assertEqual(ps[1].item.name, 'gamma')
        self.assertEqual(ps[1].item.item_type, 'P')
        self.assertTrue(ps[1].item.has_particle)
        gamma = ps[1].item.particle
        self.assertIsInstance(gamma, PdgParticle)
        self.assertEqual(gamma.pdgid, 'S000/%s' % self.api.default_edition)



"""
PDG API top-level class.
"""

import sqlalchemy
from sqlalchemy import func, select, bindparam, distinct, desc
import pdg
from pdg.errors import PdgAmbiguousValueError, PdgInvalidPdgIdError, PdgNoDataError
from pdg.utils import parse_id
from pdg.data import PdgProperty, PdgMass, PdgWidth, PdgLifetime, PdgText
from pdg.decay import PdgBranchingFraction, PdgBranchingRatio, PdgItem
from pdg.particle import PdgParticle, PdgParticleList


# Map PDG data type codes to corresponding classes
DATA_TYPE_MAP = {
    'PART': PdgParticleList,
    'M':    PdgMass,
    'BFX':  PdgBranchingFraction,
    'BFX1': PdgBranchingFraction,
    'BFX2': PdgBranchingFraction,
    'BFX3': PdgBranchingFraction,
    'BFX4': PdgBranchingFraction,
    'BFX5': PdgBranchingFraction,
    'BFI':  PdgBranchingFraction,
    'BFI1': PdgBranchingFraction,
    'BFI2': PdgBranchingFraction,
    'BFI3': PdgBranchingFraction,
    'BFI4': PdgBranchingFraction,
    'BFI5': PdgBranchingFraction,
    'BR': PdgBranchingRatio,
    'G': PdgWidth,
    'T': PdgLifetime,
    'SEC': PdgText,
}


class PdgApi:

    def __init__(self, database_url, pedantic=False):
        """Initialize PDG API.

        database_url is the URL of the PDG database to connect to. The default database is the SQLite file
        installed together with package pdg.

        pedantic can be set True to enable pedantic mode, where, in cases where the choice of "PDG best value" might
        be ambiguous, no assumptions are made and instead a PdgAmbiguousValue exception is raised.
        """
        self.database_url = database_url
        self.engine = sqlalchemy.create_engine(self.database_url)
        self.db = sqlalchemy.MetaData()
        self.db.reflect(self.engine)
        for k in self.info_keys():
            setattr(self, k, self.info(k))
        self.pedantic = pedantic

    def __str__(self):
        s = ['%s Review of Particle Physics, data release %s, API version %s' % (self.info('edition'),
                                                                                 self.info('data_release_timestamp'),
                                                                                 pdg.__version__),
             '%s' % self.info('citation'),
             '(C) %s, data released under %s' % (self.info('producer'), self.info('license')),
             self.info('about')
             ]
        return '\n'.join(s)

    def info(self, key):
        """Return metadata info specified by key."""
        pdginfo_table = self.db.tables['pdginfo']
        query = select(pdginfo_table.c.value).where(pdginfo_table.c.name == bindparam('key'))
        with self.engine.connect() as conn:
            return conn.execute(query, {'key': key}).scalar()

    def info_keys(self):
        """Return list of all metadata keys."""
        pdginfo_table = self.db.tables['pdginfo']
        query = select(pdginfo_table.c.name)
        with self.engine.connect() as conn:
            return [k[0] for k in conn.execute(query).fetchall()]

    @property
    def editions(self):
        """List of all editions of the Review for which the database has data."""
        pdgdata_table = self.db.tables['pdgdata']
        query = select(distinct(pdgdata_table.c.edition)).order_by(desc(pdgdata_table.c.edition))
        with self.engine.connect() as conn:
            return [e[0] for e in conn.execute(query).fetchall()]

    @property
    def default_edition(self):
        """Return the default edition for this database."""
        return self.info('edition')

    def get(self, pdgid, edition=None):
        """Return PdgData object for given PDG Identifier.

        The get method checks what data the PDG Identifier describes and returns an
        object of the most appropriate class derived from the PdgData base class.
        For example, for a PDG Identifier describing a particle, an object of class
        PdgParticleList is returned, while for a branching fraction a PdgBranchingFraction
        object is returned.

        edition can be set to a specific edition, from which the data should later be retrieved.
        """
        if edition is None:
            baseid, edition = parse_id(pdgid)
        else:
            baseid = pdgid
        pdgid_table = self.db.tables['pdgid']
        try:
            query = select(pdgid_table.c.data_type).where(pdgid_table.c.pdgid == bindparam('pdgid'))
            with self.engine.connect() as conn:
                data_type = conn.execute(query, {'pdgid': baseid}).fetchone()[0]
        except Exception:
            raise PdgInvalidPdgIdError('PDG Identifier %s not found' % pdgid)
        try:
            cls = DATA_TYPE_MAP[data_type]
        except KeyError:
            cls = PdgProperty
        return cls(self, baseid, edition)

    def get_all(self, data_type_key=None, edition=None):
        """Return iterator over all PDG Identifiers / quantities. Returns PdgProperties or derived classes.

        If data_type_key is set, only quantities of the given type are returned.
        See doc_data_type_keys() for the list of possible data type codes.

        edition can be set to a specific edition, from which data should later be retrieved.
        """
        pdgid_table = self.db.tables['pdgid']
        query = select(pdgid_table.c.pdgid, pdgid_table.c.data_type)
        if data_type_key is not None:
            query = query.where(pdgid_table.c.data_type == bindparam('data_type_key'))
        query = query.order_by(pdgid_table.c.sort)
        with self.engine.connect() as conn:
            for item in conn.execute(query, {'data_type_key': data_type_key}):
                try:
                    cls = DATA_TYPE_MAP[item.data_type]
                except KeyError:
                    cls = PdgProperty
                yield cls(self, item.pdgid, edition)

    def _get_particles_by_name(self, name, case_sensitive=True, edition=None, unique=True):
        """Helper function used by get_particle(s)_by_name. Returns a
        PdgParticle (list thereof) if unique is True (False). Raises a
        PdgAmbiguousValueError if more than one PdgItem exists with the given
        name, or if unique=True and the PdgItem refers to more than one
        particle.
        """
        pdgitem_table = self.db.tables['pdgitem']
        query = select(pdgitem_table.c.id)
        if case_sensitive:
            query = query.where(pdgitem_table.c.name == bindparam('name'))
        else:
            name = name.lower()
            query = query.where(func.lower(pdgitem_table.c.name) == bindparam('name'))
        with self.engine.connect() as conn:
            matches = conn.execute(query, {'name': name}).fetchall()
        if len(matches) == 0:
            raise ValueError('No particle found with name %s' % name)
        elif len(matches) == 1:
            item = PdgItem(self, matches[0].id, edition=edition)
            return item.particle if unique else item.particles
        else:
            raise PdgAmbiguousValueError('More than one PDGITEM named %s', name)

    def get_particle_by_name(self, name, case_sensitive=True, edition=None):
        """Get particle by its name.

        case_sensitive can be set False to indicate that the particle name should be
        considered not case-sensitive.

        edition can be set to a specific edition, from which data should later be retrieved.
        """
        return self._get_particles_by_name(name, case_sensitive=case_sensitive,
                                           edition=edition, unique=True)

    def get_particles_by_name(self, name, case_sensitive=True, edition=None):
        """Get all particles for a (possibly generic) name.

        case_sensitive can be set False to indicate that the particle name should be
        considered not case-sensitive.

        edition can be set to a specific edition, from which data should later be retrieved.
        """
        return self._get_particles_by_name(name, case_sensitive=case_sensitive,
                                           edition=edition, unique=False)

    def get_particle_by_mcid(self, mcid, edition=None):
        """Get particle by its MC ID.

        edition can be set to a specific edition, from which data should later be retrieved.
        """
        pdgparticle_table = self.db.tables['pdgparticle']
        query = select(distinct(pdgparticle_table.c.pdgid))
        query = query.where(pdgparticle_table.c.mcid == bindparam('mcid'))
        with self.engine.connect() as conn:
            matches = [p.pdgid for p in conn.execute(query, {'mcid': mcid})]
        if len(matches) == 0:
            raise ValueError('No particle found with MC ID %s' % mcid)
        elif len(matches) == 1:
            return PdgParticle(self, matches[0], edition, set_mcid=mcid)
        else:
            raise ValueError('MC number %s matches %i particles with PDG Identifiers %s' % (mcid, len(matches), matches))

    def get_particles(self, edition=None):
        """Return iterator over all particles.

        edition can be set to a specific edition, from which data should later be retrieved.
        """
        pdgid_table = self.db.tables['pdgid']
        pdgparticle_table = self.db.tables['pdgparticle']
        query = select(distinct(pdgid_table.c.pdgid)).join(pdgparticle_table)
        query = query.where(pdgid_table.c.data_type == 'PART')
        query = query.order_by(pdgid_table.c.sort)
        with self.engine.connect() as conn:
            for item in conn.execute(query):
                yield PdgParticleList(self, item.pdgid, edition)

    def get_canonical_name(self, name):
        return self.get_particle_by_name(name).name

    def doc_key_value(self, table_name, column_name, key):
        """Get documentation on the meaning of key values or flags used in the PDG API."""
        pdgdoc_table = self.db.tables['pdgdoc']
        query = select(pdgdoc_table)
        query = query.where(pdgdoc_table.c.table_name == bindparam('table_name'))
        query = query.where(pdgdoc_table.c.column_name == bindparam('column_name'))
        query = query.where(pdgdoc_table.c.value == bindparam('value'))
        with self.engine.connect() as conn:
            try:
                return conn.execute(query, {'table_name': table_name, 'column_name': column_name, 'value': key}).\
                    fetchone()._mapping
            except AttributeError:
                raise PdgNoDataError('No documentation for value %s in table %s.%s' % (key, table_name, column_name))

    def doc_data_type_keys(self, as_text=True):
        """Get list of data type keys.

        The PDG API uses a data type key as part of the PDG Identifier metadata to denote the kind of information
        described by a given identifier. These data type keys can be used to select desired particle properties in
        methods such as PdgParticle.properties().

        doc_data_type_keys() returns a list of all possible data type key values.
        When as_text is True (default), the list is returned as a formatted string suitable for printing.
        Otherwise, a list of dict is returned, where each dict describes a possible key value.
        """
        keys = []
        if as_text:
            keys.append('Key value     Description')
            keys.append('-'*60)
        pdgdoc_table = self.db.tables['pdgdoc']
        query = select(pdgdoc_table)
        query = query.where(pdgdoc_table.c.table_name == 'PDGID')
        query = query.where(pdgdoc_table.c.column_name == 'DATA_TYPE')
        query.order_by(pdgdoc_table.c.indicator, pdgdoc_table.c.value)
        with self.engine.connect() as conn:
            for item in conn.execute(query):
                if as_text:
                    keys.append('  %-8s    %s' % (item.value, item.description))
                else:
                    keys.append(item._mapping)
        if as_text:
            return '\n'.join(keys)
        else:
            return keys

    def doc_value_type_keys(self, as_text=True):
        """Get list of summary value type keys.

        For each summary value, the value type key specifies how this value was derived, e.g. whether it is the
        result of a weighted average, of a fit, etc.

        doc_summary_value_type_keys() returns a list of all possible summary value type key values.
        When as_text is True (default), the list is returned as a formatted string suitable for printing.
        Otherwise, a list of dict is returned, where each dict describes a possible key value.
        """
        keys = []
        if as_text:
            keys.append('Key value   Indicator            Description')
            keys.append('-'*60)
        pdgdoc_table = self.db.tables['pdgdoc']
        query = select(pdgdoc_table)
        query = query.where(pdgdoc_table.c.table_name == 'PDGDATA')
        query = query.where(pdgdoc_table.c.column_name == 'VALUE_TYPE')
        query.order_by(pdgdoc_table.c.indicator, pdgdoc_table.c.value)
        with self.engine.connect() as conn:
            for item in conn.execute(query):
                if as_text:
                    keys.append('  %-8s  %-20s  %s' % (item.value, item.indicator, item.description))
                else:
                    keys.append(item._mapping)
        if as_text:
            return '\n'.join(keys)
        else:
            return keys



"""
Python API to access PDG data.

The Python package pdg provides access to the particle physics data
published by the Particle Data Group (PDG) in the Review of Particle Physics.

For documentation see https://pdgapi.lbl.gov/doc.

For general information about PDG and the Review of Particle Physics
please visit https://pdg.lbl.gov.
"""
__author__ = 'Particle Data Group'
__version__ = '0.2.0'


import os
from pdg.api import PdgApi
from pdg.errors import PdgApiError


# Constants
SQLITE_FILENAME = 'pdg.sqlite'      # Default SQLite database file used by this API
MIN_SCHEMA_VERSION = 0.3            # Minimum schema version required by this version of the API


def connect(database_url=None, pedantic=False):
    """Connect to PDG database and return configured PDG API object."""
    if database_url is None:
        api = PdgApi('sqlite:///%s' % os.path.join(os.path.dirname(__file__), SQLITE_FILENAME), pedantic)
    else:
        api = PdgApi(database_url, pedantic)
    schema_version = float(api.info('schema_version'))
    if schema_version < MIN_SCHEMA_VERSION:
        raise PdgApiError('database schema v%s too old - need at least v%s' % (schema_version, MIN_SCHEMA_VERSION))
    return api



"""
Test cases for some use examples.
"""
from __future__ import print_function

import unittest

import pdg


class TestData(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.api = pdg.connect(pedantic=False)

    def test_dm(self):
        p = self.api.get_particle_by_name('p')
        n = self.api.get_particle_by_name('n')
        dm = n.mass - p.mass
        self.assertEqual(round(dm, 5), 0.00129)



"""
PDG API exception classes.
"""


class PdgApiError(Exception):
    """PDG API base exception."""
    pass


class PdgInvalidPdgIdError(Exception):
    """Exception raised when encountering an invalid PDG Identifier."""
    pass


class PdgNoDataError(Exception):
    """Exception raised if no data is found."""
    pass


class PdgAmbiguousValueError(Exception):
    """Exception raised in cases where the choice of value is ambiguous and there is no single best value."""
    pass


class PdgRoundingError(Exception):
    """Exception raised in cases where PDG rounding is undefined."""
    pass



"""
Test cases for the PDG data.
"""
from __future__ import print_function

import unittest

import pdg
from pdg.errors import PdgAmbiguousValueError, PdgInvalidPdgIdError
from pdg.data import PdgConvertedValue, PdgMass
from pdg.particle import PdgParticle, PdgParticleList
from pdg.decay import PdgBranchingFraction


class TestData(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.api = pdg.connect()

    def test_exception(self):
        self.assertRaises(PdgInvalidPdgIdError, self.api.get, 'nonexistent')

    def test_edition(self):
        self.assertEqual(self.api.get('s008').edition, self.api.default_edition)
        if '2018' not in self.api.editions:
            return
        self.assertEqual(self.api.get('s008/2018').pdgid, 'S008/2018')
        self.assertEqual(self.api.get('s008/2018').baseid, 'S008')
        self.assertEqual(self.api.get('s008/2018').get_parent_pdgid(), None)
        self.assertEqual(self.api.get('s008.1/2018').get_parent_pdgid(), 'S008/2018')

    def test_charged_pion_2022(self):
        if '2022' not in self.api.editions:
            return
        pion = self.api.get('S008/2022')[0]
        self.assertIsInstance(pion, PdgParticle)
        self.assertEqual(pion.pdgid, 'S008/2022')
        self.assertEqual(pion.edition, '2022')
        self.assertEqual(pion.description, 'pi+-')
        self.assertEqual(pion.data_type, 'PART')
        self.assertEqual(pion.get_parent_pdgid(), None)
        self.assertEqual(pion.name, 'pi+')
        self.assertEqual(pion.mcid, 211)
        self.assertEqual(pion.charge, 1.0)
        self.assertEqual(pion.quantum_I, '1')
        self.assertEqual(pion.quantum_G, '-')
        self.assertEqual(pion.quantum_J, '0')
        self.assertEqual(pion.quantum_P, '-')
        self.assertEqual(pion.quantum_C, None)
        self.assertEqual(sum(1 for _ in pion.properties(require_summary_data=False)), 14)
        self.assertEqual(sum(1 for _ in pion.properties(require_summary_data=True)), 10)
        self.assertEqual(sum(1 for _ in pion.exclusive_branching_fractions()), 9)
        self.assertEqual(sum(1 for _ in pion.exclusive_branching_fractions(include_subdecays=True)), 11)

    def test_charged_pion_mass_2022(self):
        if '2022' not in self.api.editions:
            return
        m = self.api.get('S008M/2022')
        self.assertEqual(m._count_data_entries('S008m', 2022), 2)
        self.assertIsInstance(m, PdgMass)
        self.assertEqual(m.description, 'pi+- MASS')
        self.assertEqual(m.n_summary_table_values(), 1)
        self.assertEqual(m.get_parent_pdgid(), 'S008/2022')
        self.assertRaises(PdgAmbiguousValueError, lambda: m.get_particle())
        self.assertIsInstance(m.get_particles(), PdgParticleList)
        self.assertIsInstance(m.get_particles()[0], PdgParticle)
        # Enable following tests only after IN_SUMMARY_TABLE is properly set
        best_value = m.best_summary()
        self.assertEqual(best_value.in_summary_table, True)
        self.assertEqual(round(best_value.value, 5), 139.57039)
        self.assertEqual(round(best_value.error_positive, 5), 0.00018)
        self.assertEqual(round(best_value.error_negative, 5), 0.00018)
        self.assertEqual(round(best_value.error,5), 0.00018)
        self.assertEqual(round(best_value.scale_factor, 1), 1.8)
        self.assertEqual(best_value.units, 'MeV')
        # self.assertEqual(best_value.units_tex, 'MeV')
        self.assertEqual(best_value.value_text, '139.57039+-0.00018')
        # self.assertEqual(best_value.value_tex, None) # FIXME
        self.assertEqual(best_value.display_value_text, '139.57039+-0.00018')
        self.assertEqual(best_value.value_type_key, 'FC')
        self.assertEqual(str(best_value).strip(), '139.57039+-0.00018   OUR FIT')
        best_value_GeV = PdgConvertedValue(best_value, 'GeV')
        self.assertEqual(round(best_value_GeV.value, 8), 0.13957039)
        self.assertEqual(round(best_value_GeV.error_positive, 8), 1.8E-7)
        self.assertEqual(round(best_value_GeV.error_negative, 8), 1.8E-7)
        self.assertEqual(round(best_value_GeV.scale_factor, 1), 1.8)
        self.assertEqual(best_value_GeV.units, 'GeV')

    def test_neutral_pion_decay_2022(self):
        if '2022' not in self.api.editions:
            return
        b = self.api.get('S009.1/2022')
        self.assertIsInstance(b, PdgBranchingFraction)
        self.assertEqual(b.description, 'pi0 --> 2gamma')
        self.assertEqual(b.n_summary_table_values(), 1)
        self.assertEqual(b.get_parent_pdgid(), 'S009/2022')
        self.assertIsInstance(b.get_particle(), PdgParticle)
        self.assertEqual(b.is_subdecay, False)
        self.assertEqual(b.subdecay_level, 0)
        best_value = b.best_summary()
        self.assertEqual(best_value.in_summary_table, True)
        self.assertEqual(round(best_value.value, 5), 0.98823)
        self.assertEqual(round(best_value.error_positive, 5), 0.00034)
        self.assertEqual(round(best_value.error_negative, 5), 0.00034)
        self.assertEqual(round(best_value.scale_factor, 1), 1.5)
        self.assertEqual(best_value.units, '')
        # self.assertEqual(best_value.units_tex, '')
        self.assertEqual(best_value.value_text, '(98.823+-0.034)E-2')
        # self.assertEqual(best_value.value_tex, None)                # FIXME
        self.assertEqual(best_value.display_value_text, '98.823+-0.034')
        self.assertEqual(best_value.display_power_of_ten, -2)
        self.assertEqual(best_value.value_type_key, 'FC')
        self.assertEqual(best_value.display_in_percent, True)

    def test_subdecay_modes(self):
        self.assertEqual(self.api.get('S008.1').is_subdecay, False)
        self.assertEqual(self.api.get('S008.3').is_subdecay, True)
        self.assertEqual(self.api.get('S008.1').subdecay_level, 0)
        self.assertEqual(self.api.get('S008.3').subdecay_level, 1)

    def test_limit_2022(self):
        if '2022' not in self.api.editions:
            return
        self.assertEqual(self.api.get('S008.11').best_summary().is_limit, True)
        self.assertEqual(self.api.get('S008.11').best_summary().is_upper_limit, True)
        self.assertEqual(self.api.get('S008.11').best_summary().is_lower_limit, False)
        self.assertEqual(self.api.get('S008FV').best_summary().is_limit, False)
        self.assertEqual(self.api.get('S008FV').best_summary().is_upper_limit, False)
        self.assertEqual(self.api.get('S008FV').best_summary().is_lower_limit, False)

    def test_error_2022(self):
        if '2022' not in self.api.editions:
            return
        self.assertEqual(self.api.get('S066M2E/2022').best_summary().error, None)
        self.assertEqual(self.api.get('M013.4/2022').best_summary().error, 0.0218379495175522 )

    def test_data_flags(self):
        self.assertEqual(self.api.get('S003AMU').data_flags, 'As')
        self.assertEqual(self.api.get('S004AMU').data_flags, 'As')
        self.assertEqual(self.api.get('S016AMU').data_flags, 'As')
        self.assertEqual(self.api.get('S017AMU').data_flags, 'A0s')
        self.assertEqual(self.api.get('Q007TP').data_flags, 'D')
        self.assertEqual(self.api.get('Q007TP2').data_flags, 's')
        self.assertEqual(self.api.get('Q007TP4').data_flags, 's')

    def test_old_bugs(self):
        # Check fix for metadata bug in v0.0.5
        self.assertIsNotNone(self.api.get('S086DRA').best_summary())
        self.assertIsNotNone(self.api.get('S086DGS').best_summary())
        # Check that LIMIT_TYPE is NULL instead of the empty string
        self.assertFalse(self.api.get('S041.3').is_limit)

    def test_value_parsing(self):
        # Check that DISPLAY_VALUE_TEXT is being parsed properly
        # Symmetric errors:
        self.assertEqual(round(self.api.get('S086T').value*1e12, 2), 1.52)
        self.assertEqual(round(self.api.get('S086T').error_positive*1e15, 1), 6)
        self.assertEqual(round(self.api.get('S086T').error_negative*1e15, 1), 6)
        self.assertEqual(round(self.api.get('S086T').error*1e15, 1), 6)
        # Slightly asymmetric errors:
        self.assertEqual(round(self.api.get('S063T').value*1e12, 2), 1.64)
        self.assertEqual(round(self.api.get('S063T').error_positive*1e13, 1), 1.6)
        self.assertEqual(round(self.api.get('S063T').error_negative*1e13, 1), 1.6)
        self.assertEqual(round(self.api.get('S063T').error*1e13, 2), 1.6)
        # Very asymmetric errors:
        self.assertEqual(round(self.api.get('S041P63').value*1e3, 1), 10.2)
        self.assertEqual(round(self.api.get('S041P63').error_positive*1e3, 1), 3.2)
        self.assertEqual(round(self.api.get('S041P63').error_negative*1e3, 1), 6.9)
        self.assertIsNone(self.api.get('S041P63').error)

    def test_HFLAV_comment(self):
        if int(self.api.default_edition) >= 2024:
            self.assertEqual(self.api.get('S042T').comment,
                             '(Produced by HFLAV)')

    def test_parent_pdgid(self):
        mass = self.api.get('Q007TP4')
        # We want to verify that the intermediate section gets skipped, so
        # first, we check that the direct parent is in fact such a section
        direct_parent = self.api.get(mass._get_pdgid()['parent_pdgid'])
        self.assertEqual(direct_parent.data_type, 'SEC')
        # Now we can check that the section gets skipped
        self.assertEqual(mass.get_parent_pdgid(include_edition=False), 'Q007')

    def test_decay_pdgid_map(self):
        bf = self.api.get('B000.2')
        br_pdgids = sorted([br.baseid for br in bf.branching_ratios()])
        self.assertEqual(br_pdgids, ['B000R01', 'B000R2'])
        br = self.api.get('B000R2')
        bf_pdgids = sorted([bf.baseid for bf in br.branching_fractions()])
        self.assertEqual(bf_pdgids, ['B000.1', 'B000.2'])


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3

from pdg.errors import PdgAmbiguousValueError
from pdg.utils import get_linked_ids, get_row_data

class PdgMeasurement(object):
    """Class for an individual measurement from the PDG Listings."""

    def __init__(self, api, msmt_id):
        self.api = api
        self.id = msmt_id
        self.cache = {}

    def _get_measurement_data(self):
        return self.cache.get(
            'pdgmeasurement',
            get_row_data(self.api, 'pdgmeasurement', self.id))

    def values(self):
        """Returns a iterator of PdgValues for all of the quanities associated
           with this measurement."""
        for value_id in get_linked_ids(
                self.api,
                'pdgmeasurement_values',
                'pdgmeasurement_id',
                self.id):
            yield PdgValue(self.api, value_id)

    def get_value(self):
        """Returns the PdgValue associated with this measurement. Raises
        PdgAmbiguousValueError if there are multiple values (i.e. this is a
        multi-column measurement)."""
        vs = list(self.values())
        if len(vs) != 1:
            err = 'Multiple values associated with this measurement. ' \
                + 'Use the values() method instead.'
            raise PdgAmbiguousValueError(err)
        return vs[0]

    def footnotes(self):
        """Returns an iterator of PdgFootnotes for this measurements."""
        for foot_id in get_linked_ids(
                self.api,
                'pdgmeasurement_footnote',
                'pdgmeasurement_id',
                self.id,
                'pdgfootnote_id'):
            yield PdgFootnote(self.api, foot_id)

    @property
    def reference(self):
        """The PdgReference associated with this measurement."""
        ref_id = self._get_measurement_data()['pdgreference_id']
        return PdgReference(self.api, ref_id)

    @property
    def pdgid(self):
        """The PDG Identifier to which this measurement applies."""
        return self._get_measurement_data()['pdgid']

    @property
    def event_count(self):
        """The number of events in the sample used for this measurement."""
        return self._get_measurement_data()['event_count']

    @property
    def confidence_level(self):
        """The confidence level of this measurement."""
        return self._get_measurement_data()['confidence_level']

    @property
    def technique(self):
        """The technique used in this measurement."""
        return self._get_measurement_data()['technique']

    @property
    def charge(self):
        """Case-specific representation of the charge(s) of the particles
           involved in this measurement."""
        return self._get_measurement_data()['charge']

    @property
    def changebar(self):
        """True if this measurement has been added or updated since the previous
           PDG edition"""
        return self._get_measurement_data()['changebar']

    @property
    def comment(self):
        """Inline comment displayed with this measurement in the PDG Listings."""
        return self._get_measurement_data()['comment']


class PdgValue(object):
    """Class for an individual numerical value associated with a PdgMeasurement."""

    def __init__(self, api, value_id):
        self.api = api
        self.id = value_id
        self.cache = {}

    def _get_value_data(self):
        return self.cache.get(
            'pdgmeasurement_values',
            get_row_data(self.api, 'pdgmeasurement_values', self.id))

    @property
    def measurement(self):
        """The corresponding PdgMeasurement for this value."""
        msmt_id = self._get_value_data()['pdgmeasurement_id']
        return PdgMeasurement(self.api, msmt_id)

    @property
    def column_name(self):
        """The name of the column (in plain text format) in which this value is
           displayed in the PDG Listings."""
        return self._get_value_data()['column_name']

    @property
    def column_name_tex(self):
        """The name of the column (in TeX format) in which this value is
        displayed in the PDG Listings."""
        return self._get_value_data()['column_name_tex']

    @property
    def unit_text(self):
        """The units (in plain text format) in which this value is specified."""
        return self._get_value_data()['unit_text']

    # @property
    # def unit_tex(self):
    #     """The units (in TeX format) in which this value is specified."""
    #     return self._get_value_data()['unit_tex']

    @property
    def value_text(self):
        """Value and uncertainty (in plain text format) in units given by
           property units, including the power of ten, if applicable
           (see display_power_of_ten)"""
        return self._get_value_data()['value_text']

    # @property
    # def value_tex(self):
    #     """Value and uncertainty (in TeX format) in units given by property
    #        units, including the power of ten, if applicable (see
    #        display_power_of_ten)"""
    #     return self._get_value_data()['value_tex']

    @property
    def display_value_text(self):
        """Value and uncertainty in plain text format as displayed in
           Listings tables. Does not include any power of ten or percent sign.
           Must be combined with the display_power_of_ten property in order
           to obtain the numerical value in units given by the property units."""
        return self._get_value_data()['display_value_text']

    @property
    def display_power_of_ten(self):
        """Unit multiplier (as power of ten) as used for display in Listings."""
        return self._get_value_data()['display_power_of_ten']

    @property
    def display_in_percent(self):
        """True if value is rendered in percent for display in Listings.
           Implies that display_power_of_ten is -2."""
        return self._get_value_data()['display_in_percent']

    @property
    def is_limit(self):
        """True if value is a limit."""
        return self._get_value_data()['limit_type'] is not None

    @property
    def is_upper_limit(self):
        """True if value is an upper limit."""
        return self._get_value_data()['limit_type'] == 'U'

    @property
    def is_lower_limit(self):
        """True if value is a lower limit."""
        return self._get_value_data()['limit_type'] == 'L'

    @property
    def used_in_average(self):
        """True if value is used for calculating averages shown in
           Summary Tables."""
        return self._get_value_data()['used_in_average']

    @property
    def used_in_fit(self):
        """True if value is used for calculating best-fits shown in
           Summary Tables."""
        return self._get_value_data()['used_in_fit']

    @property
    def value(self):
        """The numerical value itself."""
        return self._get_value_data()['value']

    @property
    def error_positive(self):
        """The total positive error for this value."""
        return self._get_value_data()['error_positive']

    @property
    def error_negative(self):
        """The total negative error for this value."""
        return self._get_value_data()['error_negative']

    @property
    def stat_error_positive(self):
        """The statistical component of the positive error for this value."""
        return self._get_value_data()['stat_error_positive']

    @property
    def stat_error_negative(self):
        """The statistical component of the negative error for this value."""
        return self._get_value_data()['stat_error_negative']

    @property
    def syst_error_positive(self):
        """The systematic component of the positive error for this value."""
        return self._get_value_data()['syst_error_positive']

    @property
    def syst_error_negative(self):
        """The systematic component of the negative error for this value."""
        return self._get_value_data()['syst_error_negative']

    @property
    def error(self):
        """The total symmetric error for this value. If the positive and
           negative errors differ, accessing this property will give the average
           of the two, unless you are in pedantic mode, in which case a
           PdgAmibguousValueError will be raised."""
        is_asymm = self.error_positive != self.error_negative
        if self.api.pedantic and is_asymm:
            msg = 'In pedantic mode, PdgValue.error is ill-defined when the ' \
                + 'postive and negative errors are unequal.'
            raise PdgAmbiguousValueError(msg)
        if self.error_positive is None or self.error_negative is None:
            return None
        return 0.5*(self.error_positive + self.error_negative)

    @property
    def stat_error(self):
        """The statistical symmetric error for this value. If the positive and
           negative statistical errors differ, accessing this property will give
           the average of the two, unless you are in pedantic mode, in which case a
           PdgAmibguousValueError will be raised."""
        is_asymm = self.stat_error_positive != self.stat_error_negative
        if self.api.pedantic and is_asymm:
            msg = 'In pedantic mode, PdgValue.stat_error is ill-defined when the ' \
                + 'postive and negative statistical errors are unequal.'
            raise PdgAmbiguousValueError(msg)
        if self.stat_error_positive is None or self.stat_error_negative is None:
            return None
        return 0.5*(self.stat_error_positive + self.stat_error_negative)

    @property
    def syst_error(self):
        """The systematic symmetric error for this value. If the positive and
           negative systematic errors differ, accessing this property will give
           the average of the two, unless you are in pedantic mode, in which case a
           PdgAmibguousValueError will be raised."""
        is_asymm = self.syst_error_positive != self.syst_error_negative
        if self.api.pedantic and is_asymm:
            msg = 'In pedantic mode, PdgValue.stat_error is ill-defined when the ' \
                + 'postive and negative systematic errors are unequal.'
            raise PdgAmbiguousValueError(msg)
        if self.syst_error_positive is None or self.syst_error_negative is None:
            return None
        return 0.5*(self.syst_error_positive + self.syst_error_negative)


class PdgReference(object):
    """Class for literature references associated with PdgMeasurements."""

    def __init__(self, api, ref_id):
        self.api = api
        self.id = ref_id
        self.cache = {}

    def _get_reference_data(self):
        return self.cache.get(
            'pdgreference',
            get_row_data(self.api, 'pdgreference', self.id))

    @property
    def publication_name(self):
        """The abbreviated bibliographic name (e.g. journal initials, issue,
           page) of this publication."""
        return self._get_reference_data()['publication_name']

    @property
    def publication_year(self):
        """The year of this publication."""
        return self._get_reference_data()['publication_year']

    @property
    def title(self):
        """The title of this publication."""
        return self._get_reference_data()['title']

    @property
    def doi(self):
        """The DOI of this publication."""
        return self._get_reference_data()['doi']

    @property
    def inspire_id(self):
        """The INSPIRE identifier of this publication."""
        return self._get_reference_data()['inspire_id']

    @property
    def document_id(self):
        """The identifier for this publication in AUTHOR YEAR format."""
        return self._get_reference_data()['document_id']


class PdgFootnote(object):
    """Class for footnotes associated with PdgReferences."""

    def __init__(self, api, foot_id):
        self.api = api
        self.id = foot_id
        self.cache = {}

    def _get_footnote_data(self):
        return self.cache.get(
            'pdgfootnote',
            get_row_data(self.api, 'pdgfootnote', self.id))

    def references(self):
        """Returns an iterator of PdgReferences that refer to this footnote."""
        for ref_id in get_linked_ids(
                self.api,
                'pdgmeasurement_footnote',
                'pdgfootnote_id',
                self.id,
                'pdgmeasurement_id'):
            yield PdgMeasurement(self.api, ref_id)

    @property
    def text(self):
        """The footnote text, in plain text format."""
        return self._get_footnote_data()['text']

    # @property
    # def text_tex(self):
    #     """The footnote text, in plain text format."""
    #     return self._get_footnote_data()['text_tex']



"""
Constants and utilities for handling HEP units.
"""

from pdg.errors import PdgApiError

HBAR_IN_GEV_S = 6.582E-25

UNIT_CONVERSION_FACTORS = {
    'meV': (1E-3, 'eV'),
    'eV': (1E0, 'eV'),
    'keV': (1E3, 'eV'),
    'MeV': (1E6, 'eV'),
    'GeV': (1E9, 'eV'),
    'TeV': (1E12, 'eV'),
    'PeV': (1E15, 'eV'),
    'u': (931.49410242E6, 'eV'),
    's': (1E0, 's'),
    'yr': (31536000, 's'),
    'year': (31536000, 's'),
    'years': (31536000, 's'),
}


def convert(value, old_units=None, new_units=None):
    """Utility to convert value to a different unit."""
    if new_units is None:
        return value
    else:
        try:
            old_factor = UNIT_CONVERSION_FACTORS[old_units]
        except KeyError:
            raise PdgApiError('Cannot convert from %s' % old_units)
        try:
            new_factor = UNIT_CONVERSION_FACTORS[new_units]
        except KeyError:
            raise PdgApiError('Cannot convert to %s' % new_units)
        if old_factor[1] != new_factor[1]:
            raise PdgApiError('Illegal unit conversion from %s to %s', old_factor[1], new_factor[1])
        return value * old_factor[0] / new_factor[0]



"""
Classes supporting decays and branching fractions/ratios.
"""

from sqlalchemy import bindparam, select

from pdg.data import PdgProperty
from pdg.errors import PdgAmbiguousValueError, PdgInvalidPdgIdError, PdgNoDataError
from pdg.particle import PdgItem, PdgParticle


class PdgDecayProduct(object):
    """Class for all information about one product of a decay, including its
    PdgItem (which may resolve to one or more PdgParticles), its multiplier, and
    its subdecay (if any).
    """
    def __init__(self, item, multiplier, subdecay):
        """Instantiate a PdgDecayProduct."""
        assert isinstance(item, PdgItem)
        assert isinstance(multiplier, int)
        assert subdecay is None or isinstance(subdecay, PdgBranchingFraction)

        self.item = item
        self.multiplier = multiplier
        self.subdecay = subdecay

    def __repr__(self):
        fmt = "PdgDecayProduct(item='%s', multiplier=%d, subdecay=%r)"
        return fmt % (self.item.name, self.multiplier, self.subdecay)


class PdgBranchingFraction(PdgProperty):
    """Class for all information about a decay, including its branching
    fraction, decay products, and subdecays.
    """
    def _get_decay(self):
        """Load decay information from the database."""
        if 'pdgdecay' not in self.cache:
            pdgdecay_table = self.api.db.tables['pdgdecay']
            query = select(pdgdecay_table).where(pdgdecay_table.c.pdgid == bindparam('pdgid'))
            with self.api.engine.connect() as conn:
                try:
                    result = conn.execute(query, {'pdgid': self.baseid}).fetchall()
                    self.cache['pdgdecay'] = [row._mapping for row in result]
                except AttributeError:
                    raise PdgInvalidPdgIdError('No PDGDECAY entry for ' % self.pdgid)
        return self.cache['pdgdecay']

    def _repr_extra(self):
        return '"%s"' % self.description

    @property
    def decay_products(self):
        """A list of all PdgDecayProducts for the decay."""
        products = []
        for row in self._get_decay():
            if not row['is_outgoing']:
                continue

            product = PdgDecayProduct(
                item=PdgItem(self.api, row['pdgitem_id']),
                multiplier=row['multiplier'],
                subdecay=(PdgBranchingFraction(self.api, row['subdecay'])
                          if row['subdecay_id'] else None))
            products.append(product)
        return products

    @property
    def mode_number(self):
        """Mode number of this decay.

        Note that the decay mode number may change from one edition of the Review of Particle Physics
        to the next one."""
        return self._get_pdgid()['mode_number']

    @property
    def is_subdecay(self):
        """True if this is a subdecay ("indented") decay mode."""
        data_type_code = self.data_type
        if len(data_type_code) < 4:
            return False
        else:
            return data_type_code[0:3] == 'BFX' or data_type_code[0:3] == 'BFI'

    @property
    def subdecay_level(self):
        """Return indentation level of a decay mode."""
        if self.is_subdecay:
            return int(self.data_type[3])
        else:
            return 0

    def branching_ratios(self):
        """Return iterator over all branching ratios associated with this
        branching fraction."""
        pdgid_map = self.api.db.tables['pdgid_map']
        query = select(pdgid_map.c.target) \
            .where(pdgid_map.c.source == bindparam('source'))
        with self.api.engine.connect() as conn:
            matches = conn.execute(query, {'source': self.baseid}).fetchall()
        for row in matches:
            yield PdgBranchingRatio(self.api, row.target, self.edition)


class PdgBranchingRatio(PdgProperty):
    """Class for all information about a branching ratio."""
    def branching_fractions(self):
        """Return iterator over all branching fractions associated with this
        branching ratio."""
        pdgid_map = self.api.db.tables['pdgid_map']
        query = select(pdgid_map.c.source) \
            .where(pdgid_map.c.target == bindparam('target'))
        with self.api.engine.connect() as conn:
            matches = conn.execute(query, {'target': self.baseid}).fetchall()
        for row in matches:
            yield PdgBranchingFraction(self.api, row.source, self.edition)

    def _repr_extra(self):
        return '"%s"' % self.description



"""
Definition of top-level particle container class.
"""

from sqlalchemy import select, bindparam, distinct, func
from sqlalchemy import and_, or_
from pdg.errors import PdgApiError, PdgNoDataError, PdgAmbiguousValueError
from pdg.utils import make_id, best
from pdg.data import PdgData
from pdg.units import HBAR_IN_GEV_S


class PdgItem:
    """A class to represent an "item" encountered in e.g. a description of a
    decay's products. An item can correspond directly to one particle,
    indirectly to one particle (as an alias), to a set of particles (as a
    generic name), or to an arbitrary string. When possible, a PdgItem can be
    queried (via the particle/particles properties) for the associated
    particle(s).
    """
    def __init__(self, api, pdgitem_id, edition=None):
        """Constructor for a PdgItem. Intended for internal API use."""
        self.api = api
        self.pdgitem_id = pdgitem_id
        self.cache = {}
        self.edition = edition

    def __repr__(self):
        """Return a human-readable representation of the PdgItem."""
        name = self._get_pdgitem()['name']
        return 'PdgItem("%s")' % name

    def _get_pdgitem(self):
        """Load the PdgItem's data from the database."""
        if 'pdgitem' not in self.cache:
            pdgitem_table = self.api.db.tables['pdgitem']
            query = select(pdgitem_table).where(pdgitem_table.c.id == bindparam('pdgitem_id'))
            with self.api.engine.connect() as conn:
                result = conn.execute(query, {'pdgitem_id': self.pdgitem_id}).fetchone()
                if result is None:
                    raise PdgNoDataError('No PDGITEM entry for %s' % self.pdgitem_id)
                self.cache['pdgitem'] = result._mapping
        return self.cache['pdgitem']

    def _get_targets(self):
        """Get all PdgItems that this one maps directly to. Does not recurse."""
        pdgitem_map_table = self.api.db.tables['pdgitem_map']
        query = select(pdgitem_map_table).where(pdgitem_map_table.c.pdgitem_id == bindparam('pdgitem_id'))
        with self.api.engine.connect() as conn:
            rows = conn.execute(query, {'pdgitem_id': self.pdgitem_id}).fetchall()
            for row in rows:
                yield PdgItem(self.api, row.target_id)

    @property
    def has_particle(self):
        """Whether the PdgItem is associated with exactly one particle. The
        property has_particles indicates whether it is associated with one or
        more.
        """
        if 'has_particle' not in self.cache:
            pdgparticle_table = self.api.db.tables['pdgparticle']
            query = select(pdgparticle_table).where(pdgparticle_table.c.pdgitem_id == bindparam('pdgitem_id'))
            with self.api.engine.connect() as conn:
                result = conn.execute(query, {'pdgitem_id': self.pdgitem_id}).fetchone()
                if result:
                    self.cache['pdgparticle'] = result._mapping
                    self.cache['has_particle'] = True
                else:
                    targets = list(self._get_targets())
                    if len(targets) == 1 and targets[0].has_particle:
                        self.cache['pdgparticle'] = targets[0].cache['pdgparticle']
                        self.cache['has_particle'] = True
                    else:
                        self.cache['has_particle'] = False
        return self.cache['has_particle']

    @property
    def particle(self):
        """The particle associated with the PdgItem, if there is exactly one
        such particle. Raises PdgAmbiguousValueError if there are more than one,
        in which case the particles property can be used instead.
        """
        if not self.has_particle:
            if self.has_particles:
                raise PdgAmbiguousValueError('No unique PDGPARTICLE for PDGITEM %s' % self.pdgitem_id)
            raise PdgNoDataError('No PDGPARTICLE for PDGITEM %s' % self.pdgitem_id)
        p = self.cache['pdgparticle']
        return PdgParticle(self.api, p['pdgid'], edition=self.edition, set_mcid=p['mcid'],
                           set_name=p['name'])

    @property
    def particles(self):
        """The list of all particles associated with the PdgItem."""
        if self.has_particle:
            return [self.particle]
        else:
            result = []
            for target in self._get_targets():
                for p in target.particles:
                    result.append(p)
            return result

    @property
    def has_particles(self):
        """Whether the PdgItem is associated with at least one particle."""
        return len(list(self.particles)) > 0

    @property
    def name(self):
        """The name of the PdgItem."""
        return self._get_pdgitem()['name']

    # @property
    # def name_tex(self):
    #     """The TeX name of the PdgItem."""
    #     return self._get_pdgitem()['name_tex']

    @property
    def item_type(self):
        """The type of the PdgItem. A single character with the following meanings:
              'P': specific state (e.g. "pi+"),
              'A': "also" alias,
              'W': "was" alias,
              'S': shortcut,
              'B': both charges (e.g. "pi+-"),
              'C': both charges, conjugate (e.g. "pi-+"),
              'G': generic state (e.g. "pi"),
              'L': general list (e.g. "leptons"),
              'I': inclusive indicator (e.g. "X"),
              'T': arbitrary text
        """
        return self._get_pdgitem()['item_type']


class PdgParticle(PdgData):
    """Container class for all information about a given particle.

    In addition to access to basic particle properties such as mass, charge, quantum numbers and
    PDG MC ID, this class provides methods to iterate over the data on all particle properties listed
    in Particle Listings and Summary Tables, including branching fractions, masses, life-times, etc.
    """

    def __init__(self, api, pdgid, edition=None, set_mcid=None, set_name=None):
        """Constructor for a PdgParticle given its PDG Identifier and possibly its MC ID."""
        super(PdgParticle, self).__init__(api, pdgid, edition)
        self.set_mcid = set_mcid
        self.set_name = set_name

    def __str__(self):
        try:
            return 'Data for PDG Particle %s: %s' % (self.pdgid, self.name)
        except PdgAmbiguousValueError:
            return 'Data for PDG Particle %s: multiple particle matches' % self.pdgid

    def _repr_extra(self):
        return "name='%s'" % self.name

    def _get_particle_data(self):
        """Get particle data."""
        if 'pdgparticle' not in self.cache:
            pdgparticle_table = self.api.db.tables['pdgparticle']
            query = select(pdgparticle_table)
            query = query.where(pdgparticle_table.c.pdgid == bindparam('pdgid'))
            if self.set_mcid is not None:
                query = query.where(pdgparticle_table.c.mcid == bindparam('mcid'))
            if self.set_name is not None:
                query = query.where(pdgparticle_table.c.name == bindparam('name'))
            with self.api.engine.connect() as conn:
                params = {'pdgid': self.baseid, 'mcid': self.set_mcid, 'name': self.set_name}
                matches = conn.execute(query, params).fetchall()
            if len(matches) == 1:
                self.cache['pdgparticle'] = matches[0]._mapping
            else:
                # Charge-specific state either not found or ambiguous
                query = select(pdgparticle_table)
                query = query.where(pdgparticle_table.c.pdgid == bindparam('pdgid'))
                if self.set_mcid is not None:
                    query = query.where(pdgparticle_table.c.mcid == bindparam('mcid'))
                query = query.where(pdgparticle_table.c.cc_type == 'S')
                query = query.where(pdgparticle_table.c.name.notlike('%bar%'))   # Exclude generic "*bar" states
                with self.api.engine.connect() as conn:
                    params = {'pdgid': self.baseid, 'mcid': self.set_mcid}
                    matches_g = conn.execute(query, params).fetchall()
                if len(matches_g) == 0:
                    mcid_string = ', MC ID = %s' % self.set_mcid if self.set_mcid else ''
                    raise PdgNoDataError('Particle data for %s%s not found' % (self.pdgid, mcid_string))
                elif len(matches_g) == 1:
                    self.cache['pdgparticle'] = matches_g[0]._mapping
                else:
                    names = [p.name for p in matches_g]
                    mcids = list(set([p.mcid for p in matches]))
                    raise PdgAmbiguousValueError('Multiple particles for %s: MCID %s, names %s' % (self.baseid, mcids, names))
        return self.cache['pdgparticle']

    def properties(self,
                   data_type_key=None,
                   require_summary_data=True,
                   in_summary_table=None,
                   omit_branching_ratios=False):
        """Return iterator over specified particle property data.

        By default, all properties excluding branching fractions and branching fraction ratios are returned.

        data_type_key can be set to select specific properties. Possible keys are given by the list printed
        by PdgApi.doc_data_type_keys. The SQL wildcard character ('%') is allowed, so to select all properties,
        including branching fractions and ratios, set data_type_key='%'. As another example, to get all mass
        properties, use data_type_key='M'.

        require_summary_data can be set False to also include properties, where the selected edition of the Review
        of Particle Physics has no summary value(s) in Particle Listings or Summary Table.

        in_summary_table can be set to select properties, where a summary value is (True) or is not (False) included
        in the Summary Table for the selected edition. Setting in_summary_table to a value other than None
        implies require_summary_data=True.

        omit_branching_ratios can be set to True to exclude any branching fraction ratio properties that would
        be selected otherwise.
        """
        pdgid_table = self.api.db.tables['pdgid']
        query = select(distinct(pdgid_table.c.pdgid))
        if require_summary_data or in_summary_table is not None:
            pdgdata_table = self.api.db.tables['pdgdata']
            query = query.join(pdgdata_table)
            query = query.where(pdgdata_table.c.edition == bindparam('edition'))
            if in_summary_table is not None:
                query = query.where(pdgdata_table.c.in_summary_table == bindparam('in_summary_table'))
        query = query.where(pdgid_table.c.parent_pdgid.like(bindparam('parent_id')))
        if data_type_key is None:
            # NOTE: like/notlike SQL operators never match null values
            query = query.where((pdgid_table.c.data_type.notlike('BF%')) | (pdgid_table.c.data_type.is_(None)))
            query = query.where((pdgid_table.c.data_type.notlike('BR%')) | (pdgid_table.c.data_type.is_(None)))
        else:
            # NOTE: like may or may not be case-sensitive, depending on database, so use it only for BR* and BF*
            # NOTE: like will not match null values, so data_type='%' must be treated separately
            if '%' in data_type_key:
                if data_type_key != '%':
                    query = query.where(pdgid_table.c.data_type.like(bindparam('data_type_key')))
            else:
                query = query.where(pdgid_table.c.data_type == bindparam('data_type_key'))
            if omit_branching_ratios:
                query = query.where((pdgid_table.c.data_type.notlike('BR%')) | (pdgid_table.c.data_type.is_(None)))
        query = query.order_by(pdgid_table.c.sort)
        with self.api.engine.connect() as conn:
            for entry in conn.execute(query, {'parent_id': self.baseid+'%',
                                              'edition': self.edition,
                                              'data_type_key': data_type_key,
                                              'in_summary_table': in_summary_table}):
                prop = self.api.get(make_id(entry.pdgid, self.edition))

                # For masses, widths, and lifetimes, we must take care to choose
                # the appropriate entry according to the particle's charge.
                # Other types of properties don't require further checks.
                if prop.data_type not in 'MGT':
                    yield prop

                # NOTE: Now that 's' properties are sorted last, we can safely
                # include them without breaking best() etc.

                # If this property is not charge-specific, yield it.
                elif not any(flag in prop.data_flags for flag in '012'):
                    yield prop

                # If this particle isn't a specific charge state, yield
                # everything.
                elif self.charge is None:
                    yield prop

                # Finally check whether the charges match
                elif str(int(abs(self.charge))) in prop.data_flags:
                    yield prop


    def masses(self, require_summary_data=True):
        """Return iterator over mass data.

        For most particles, there is only a single mass property, and so the particle's
        PDG best mass value in GeV can be obtained e.g. from

        list(some_particle.masses())[0].best_value_in_GeV()

        For other particles (e.g. for the top quark) there are different ways to determine the mass and the user
        needs to decide which mass value is the most appropriate for their use case.
        """
        return self.properties('M', require_summary_data)

    def widths(self, require_summary_data=True):
        """Return iterator over width data."""
        return self.properties('G', require_summary_data)

    def lifetimes(self, require_summary_data=True):
        """Return iterator over lifetime data."""
        return self.properties('T', require_summary_data)

    def branching_fractions(self, data_type_key='BF%', require_summary_data=True):
        """Return iterator over given type(s) of branching fraction data.

        With data_type_key='BF%' (default), all branching fractions, including subdecay modes, are returned.

        require_summary_data can be set False to include branching fractions where the current edition has no
        summary value(s) in the Particle Listings or Summary Table.
        """
        if data_type_key[0:2] != 'BF':
            raise PdgApiError('illegal branching fraction data type key %s' % data_type_key)
        return self.properties(data_type_key, require_summary_data)

    def exclusive_branching_fractions(self, include_subdecays=False, require_summary_data=True):
        """Return iterator over exclusive branching fraction data.

        Set include_subdecays to True (default is False) to also include
        subdecay modes (i.e. modes shown indented in the Summary Tables).

        require_summary_data can be set False to include branching fractions where the current edition has
        no summary value(s) in the Particle Listings or Summary Table.
        """
        if include_subdecays:
            return self.branching_fractions('BFX%', require_summary_data)
        else:
            return self.branching_fractions('BFX', require_summary_data)

    def inclusive_branching_fractions(self, include_subdecays=False, require_summary_data=True):
        """Return iterator over inclusive branching fraction data.

        Set include_subdecays to True (default is False) to also include
        subdecay modes (i.e. modes shown indented in the Summary Tables).

        require_summary_data can be set True to request only branching fractions where the current edition has
        summary value(s) in the Particle Listings or Summary Table.
        """
        if include_subdecays:
            return self.branching_fractions('BFI%', require_summary_data)
        else:
            return self.branching_fractions('BFI', require_summary_data)

    @property
    def name(self):
        """Name of particle (ASCII format)."""
        return self._get_particle_data()['name']

    @property
    def mcid(self):
        """Monte Carlo ID of particle."""
        return self._get_particle_data()['mcid']

    @property
    def charge(self):
        """Charge of particle in units of e."""
        return self._get_particle_data()['charge']

    @property
    def quantum_I(self):
        """Quantum number I (isospin) of particle."""
        return self._get_particle_data()['quantum_i']

    @property
    def quantum_G(self):
        """Quantum number G (G parity) of particle."""
        return self._get_particle_data()['quantum_g']

    @property
    def quantum_J(self):
        """Quantum number J (spin) of particle."""
        return self._get_particle_data()['quantum_j']

    @property
    def quantum_P(self):
        """Quantum number P (parity) of particle."""
        return self._get_particle_data()['quantum_p']

    @property
    def quantum_C(self):
        """Quantum number C (C parity) of particle."""
        return self._get_particle_data()['quantum_c']

    @property
    def is_boson(self):
        """True if particle is a gauge boson."""
        return 'G' in self.data_flags

    @property
    def is_quark(self):
        """True if particle is a quark."""
        return 'Q' in self.data_flags

    @property
    def is_lepton(self):
        """True if particle is a lepton."""
        return 'L' in self.data_flags

    @property
    def is_meson(self):
        """True if particle is a meson."""
        return 'M' in self.data_flags

    @property
    def is_baryon(self):
        """True if particle is a baryon."""
        return 'B' in self.data_flags

    @staticmethod
    def _if_not_limit(prop, units, error=False):
        """If a PdgProperty's best summary value is NOT a limit, return it in
        the specified units. Otherwise return None. If error is True, return the
        error instead of the value.
        """
        summary = prop.best_summary()
        if summary.is_limit:
            return None
        if error:
            return summary.get_error(units)
        return summary.get_value(units)

    @property
    def mass(self):
        """Mass of the particle in GeV."""
        best_mass_property = best(self.masses(), self.api.pedantic, '%s mass (%s)' % (self.name, self.pdgid))
        return self._if_not_limit(best_mass_property, 'GeV')

    @property
    def mass_error(self):
        """Symmetric error on mass of particle in GeV, or None if mass error are asymmetric or mass is a limit."""
        best_mass_property = best(self.masses(), self.api.pedantic, '%s (%s)' % (self.pdgid, self.description))
        return self._if_not_limit(best_mass_property, 'GeV', error=True)

    @property
    def width(self):
        """Width of the particle in GeV."""
        try:
            best_width_property = best(self.widths(), self.api.pedantic, '%s width (%s)' % (self.name, self.pdgid))
            return self._if_not_limit(best_width_property, 'GeV')
        except PdgNoDataError:
            if self.api.pedantic:
                raise
            # S063 has a lifetime entry but it's NULL
            if (not self.has_lifetime_entry) or (self.lifetime is None):
                return 0.
            return HBAR_IN_GEV_S / self.lifetime

    @property
    def width_error(self):
        """Symmetric error on width of particle in GeV, or None if width error are asymmetric or width is a limit."""
        try:
            best_width_property = best(self.widths(), self.api.pedantic, '%s (%s)' % (self.pdgid, self.description))
            return self._if_not_limit(best_width_property, 'GeV', error=True)
        except PdgNoDataError:
            if self.api.pedantic:
                raise
            # S063 has a lifetime entry but it's NULL
            if (not self.has_lifetime_entry) or (self.lifetime is None):
                return 0.
            return self.lifetime_error * HBAR_IN_GEV_S / self.lifetime**2

    @property
    def lifetime(self):
        """Lifetime of the particle in seconds."""
        try:
            best_lifetime_property = best(self.lifetimes(), self.api.pedantic, '%s lifetime (%s)' % (self.name, self.pdgid))
            return self._if_not_limit(best_lifetime_property, 's')
        except PdgNoDataError:
            if self.api.pedantic:
                raise
            if not self.has_width_entry:
                return float('inf')
            return HBAR_IN_GEV_S / self.width

    @property
    def lifetime_error(self):
        """Symmetric error on lifetime of particle in seconds, or None if lifetime error are asymmetric or lifetime is a limit."""
        try:
            best_lifetime_property = best(self.lifetimes(), self.api.pedantic, '%s (%s)' % (self.pdgid, self.description))
            err = self._if_not_limit(best_lifetime_property, 's', error=True)
            if err is None:
                err = 0.
            return err
        except PdgNoDataError:
            if self.api.pedantic:
                raise
            if not self.has_width_entry:
                return 0.
            return self.width_error * HBAR_IN_GEV_S / self.width**2

    @property
    def has_mass_entry(self):
        """Whether the particle has at least one defined mass."""
        return next(self.masses(), None) is not None

    @property
    def has_width_entry(self):
        """Whether the particle has at least one defined decay width."""
        return next(self.widths(), None) is not None

    @property
    def has_lifetime_entry(self):
        """Whether the particle has at least one defined lifetime."""
        return next(self.lifetimes(), None) is not None

    def mass_measurements(self, require_summary_data=True):
        for m in self.masses(require_summary_data=require_summary_data):
            for msmt in m.get_measurements():
                yield msmt

    def lifetime_measurements(self, require_summary_data=True):
        for t in self.lifetimes(require_summary_data=require_summary_data):
            for msmt in t.get_measurements():
                yield msmt

    def width_measurements(self, require_summary_data=True):
        for g in self.widths(require_summary_data=require_summary_data):
            for msmt in g.get_measurements():
                yield msmt


class PdgParticleList(PdgData, list):
    """A PdgData subclass to represent a list of PdgParticles. A PdgParticleList
    is returned when PdgApi.get is called with the PDGID of a (group of)
    particles.
    """
    def __init__(self, api, pdgid, edition=None):
        """Constructor for a PdgParticleList given its PDG Identifier."""
        super(PdgParticleList, self).__init__(api, pdgid, edition)

        pdgparticle_table = self.api.db.tables['pdgparticle']
        query = select(pdgparticle_table)
        query = query.where(func.lower(pdgparticle_table.c.pdgid) == bindparam('pdgid'))
        with self.api.engine.connect() as conn:
            result = conn.execute(query, {'pdgid': pdgid.lower()}).fetchall()
            for row in result:
                self.append(PdgParticle(api, pdgid, edition=edition, set_mcid=row.mcid,
                                        set_name=row.name))



# Standard package setup
from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='pdg',
    version='0.2.0',
    author='Particle Data Group',
    author_email='jberinger@lbl.gov',
    description='Python API for accessing PDG data',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://pdg.lbl.gov',
    license='Modified BSD',
    packages=find_packages(),
    package_data={"pdg": ["pdg.sqlite"]},
    install_requires=['SQLAlchemy>=1.4'],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Topic :: Database :: Front-Ends',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Physics'
        ],
    keywords='PDG, particle physics',
    project_urls={
        'Documentation': 'https://pdgapi.lbl.gov/doc/',
        'GitHub': 'https://github.com/particledatagroup/api',
        'Changelog': 'https://github.com/particledatagroup/api/blob/main/CHANGELOG.md'
    },
)



"""
Container base classes for PDG data.

All PDG data classes use lazy (i.e. only when needed) loading of data from the database
as implemented in the PdgData base class. In most cases, the data is read only once and
cached for subsequent use.

PdgProperty is a subclass of PdgData and adds the retrieval of summary values, measurement
data, etc., that is shared by all classes supporting the retrieval of different kinds of
particle physics properties such as branching fractions or particle masses.
"""

import pprint
from sqlalchemy import select, bindparam, func
from pdg.utils import parse_id, make_id
from pdg.units import UNIT_CONVERSION_FACTORS, convert
from pdg.errors import PdgApiError, PdgInvalidPdgIdError, PdgAmbiguousValueError, PdgNoDataError
from pdg.measurement import PdgMeasurement


class PdgSummaryValue(dict):
    """Container for a single value from the Summary Tables."""

    def __str__(self):
        indicator = self.value_type
        if not indicator:
            indicator = '[key = %s]' % self.value_type_key
        return '%-20s %-20s  %s' % (self.value_text, indicator, self.comment if self.comment else '')

    def pprint(self):
        """Print all data in this PdgSummaryValue object in a nice format (for debugging)."""
        pprint.pprint(self)

    def get_value(self, units=None):
        """Return value after conversion into units specified by parameter units (string).

        If units are not specified, the value is returned without conversion in the default units for this quantity.
        Check properties is_limit, is_lower_limit and is_upper_limit to determine if value is a central value or limit.
        """
        if units is None:
            return self['value']
        else:
            try:
                return convert(self['value'], self['unit_text'], units)
            except TypeError:
                return None

    def get_error_positive(self, units=None):
        """Return positive error after conversion into units specified by parameter units (string).

        If units are not specified, the positive error is returned without conversion in the default
        units for this quantity.
        """
        if units is None:
            return self['error_positive']
        else:
            return convert(self['error_positive'], self['unit_text'], units)

    def get_error_negative(self, units=None):
        """Return negative error after conversion into units specified by parameter units (string).

        If units are not specified, the negative error is returned without conversion in the default
        units for this quantity.
        """
        if units is None:
            return self['error_negative']
        else:
            return convert(self['error_negative'], self['unit_text'], units)

    def get_error(self, units=None):
        """Symmetric error or None, in units specified by parameter units (string).

        Returns symmetric error as average of positive and negative errors if they differ by less than 10% of
        their average. Otherwise, returns None. Also returns None if the quantity is a limit."""
        if self.is_limit:
            return None
        try:
            err_avg = (self.error_positive + self.error_negative) / 2.0
            if abs(self.error_positive - self.error_negative) < 0.1 * err_avg:
                return convert(err_avg, self['unit_text'], units)
            else:
                return None
        except TypeError:
            return None

    @property
    def pdgid(self):
        """PDG Identifier of quantity for which value is given."""
        return self['pdgid']

    @property
    def description(self):
        """Description of quantity for which value is given"""
        return self['description']

    @property
    def value_type_key(self):
        """Type of value, given by its key.

        See PdgApi.doc_value_type_keys() for the meaning of the different value type keys."""
        return self['value_type']

    @property
    def value_type(self):
        """Type of value, given as the PDG indicator string."""
        if self.value_type_key in ('AC', 'D', 'E'):
            return 'OUR AVERAGE'
        elif self.value_type_key in ('L',):
            return 'BEST LIMIT'
        elif self.value_type_key in ('OL',):
            return 'OUR LIMIT'
        elif self.value_type_key in ('FC', 'DR'):
            return 'OUR FIT'
        elif self.value_type_key in ('V', 'DV'):
            return 'OUR EVALUATION'
        else:
            return ''

    @property
    def in_summary_table(self):
        """True if value is included in Summary Table."""
        return self['in_summary_table']

    @property
    def confidence_level(self):
        """Confidence level for limits, None otherwise."""
        return self['confidence_level']

    @property
    def is_limit(self):
        """True if value is a limit."""
        return self['confidence_level'] is not None or self['limit_type'] is not None

    @property
    def is_upper_limit(self):
        """True if value is an upper limit."""
        return self['limit_type'] == 'U'

    @property
    def is_lower_limit(self):
        """True if value is an lower limit."""
        return self['limit_type'] == 'L'

    @property
    def comment(self):
        """Details for or comments on this value."""
        return self['comment']

    @property
    def value(self):
        """Numerical value in units given by property units.

        Check properties is_limit, is_lower_limit and is_upper_limit to determine if value is a central value or limit.
        """
        return self['value']

    @property
    def error_positive(self):
        """Numerical value of positive error in units given by property units."""
        return self['error_positive']

    @property
    def error_negative(self):
        """Numerical value of negative error in units given by property units."""

        return self['error_negative']

    @property
    def error(self):
        """Symmetric error or None.

        Returns symmetric error as average of positive and negative errors if they differ by less than 10% of
        their average. Otherwise, returns None. Also returns None if the quantity is a limit."""
        return self.get_error()

    @property
    def scale_factor(self):
        """PDG error scale factor that was applied to error_positive and error_negative."""
        return self['scale_factor'] or 1.0

    @property
    def units(self):
        """Units (in plain text format) used by value, error_positive,
        error_negative, and display_value_text."""
        return self['unit_text']

    # @property
    # def units_tex(self):
    #     """Units (in TeX format) used by value, error_positive, error_negative,
    #     and display_value_text."""
    #     return self['unit_tex']

    @property
    def value_text(self):
        """Value and uncertainty (in plain text format) in units given by
           property units, including the power of ten, if applicable
           (see display_power_of_ten)"""
        return self['value_text']

    # @property
    # def value_tex(self):
    #     """Value and uncertainty (in TeX format) in units given by property
    #        units, including the power of ten, if applicable (see
    #        display_power_of_ten)"""
    #     return self['value_tex']

    @property
    def display_value_text(self):
        """Value and uncertainty in plain text format as displayed in
           Listings tables. Does not include any power of ten or percent sign.
           Must be combined with the display_power_of_ten property in order
           to obtain the numerical value in units given by the property units."""
        return self['display_value_text']

    @property
    def display_power_of_ten(self):
        """Unit multiplier (as power of ten) as used for display in Listings."""
        return self['display_power_of_ten']

    @property
    def display_in_percent(self):
        """True if value is rendered in percent for display in Listings.
           Implies that display_power_of_ten is -2."""
        return self['display_in_percent']


class PdgConvertedValue(PdgSummaryValue):
    """A PdgSummaryValue class for storing summary values after unit conversion."""

    def __init__(self, value, to_units):
        """Instantiate a copy of PdgSummaryValue value with new units to_unit."""
        super(PdgConvertedValue, self).__init__(value)
        self.original_units = value.units
        try:
            old_factor = UNIT_CONVERSION_FACTORS[self.original_units]
        except KeyError:
            raise PdgApiError('Cannot convert from %s' % self.original_units)
        try:
            new_factor = UNIT_CONVERSION_FACTORS[to_units]
        except KeyError:
            raise PdgApiError('Cannot convert to %s' % to_units)
        if old_factor[1] != new_factor[1]:
            raise PdgApiError('Illegal unit conversion from %s to %s', old_factor[1], new_factor[1])
        conversion_factor = old_factor[0]/new_factor[0]
        for k in ('value', 'error_positive', 'error_negative', ):
            if self[k] is not None:
                self[k] *= conversion_factor
        for k in ('value_text', 'display_power_of_ten'):
            self[k] = None
        self['display_in_percent'] = False
        self['unit_text'] = to_units


class PdgData(object):
    """Base class for PDG data containers.

    This class implements the lazy data retrieval from the database
    and is the base class for all PDG data container classes.
    """

    def __init__(self, api, pdgid, edition=None):
        """Instantiate a PdgData object for the given PDG Identifier pdgid.

        When a PdgData object is instantiated, the edition of the Review of Particle Physics
        from which data will be retrieved is determined by the first edition information found
        from the following list:

        1. An edition specified as part of the PDG Identifier
        2. An edition specified by parameter edition
        3. The default edition specified by the database to which the API is connected

        The chosen edition can be queried by calling edition() and changed at any by calling set_edition().
        """
        self.api = api
        self.baseid, self._edition = parse_id(pdgid)
        if self._edition is None:
            self._edition = edition
        if self._edition is None:
            self._edition = self.api.edition
        self.pdgid = make_id(self.baseid, self._edition)
        self.cache = dict()

    def __str__(self):
        return 'Data for PDG Identifier %s: %s' % (self.pdgid, self.description)

    def __repr__(self):
        extra = self._repr_extra()
        if extra:
            extra = ', ' + extra
        return "%s('%s'%s)" % (self.__class__.__name__, make_id(self.baseid, self.edition),
                               extra)

    def _repr_extra(self):
        """A method that subclasses can override in order to add info to the
        result of __repr__.
        """
        return ''

    def _get_pdgid(self):
        """Get PDG Identifier information."""
        if 'pdgid' not in self.cache:
            pdgid_table = self.api.db.tables['pdgid']
            query = select(pdgid_table).where(pdgid_table.c.pdgid == bindparam('pdgid'))
            with self.api.engine.connect() as conn:
                try:
                    self.cache['pdgid'] = conn.execute(query, {'pdgid': self.baseid}).fetchone()._mapping
                except AttributeError:
                    raise PdgInvalidPdgIdError('PDG Identifier %s not found' % self.pdgid)
        return self.cache['pdgid']

    def _get_summary_values(self):
        """Get all summary data values."""
        if 'summary' not in self.cache:
            pdgid_table = self.api.db.tables['pdgid']
            pdgdata_table = self.api.db.tables['pdgdata']
            query = select(pdgdata_table, pdgid_table.c.description).join(pdgid_table)
            query = query.where(pdgid_table.c.pdgid == bindparam('pdgid'))
            query = query.where(pdgdata_table.c.edition == bindparam('edition'))
            query = query.order_by(pdgdata_table.c.sort)
            self.cache['summary'] = []
            with self.api.engine.connect() as conn:
                for entry in conn.execute(query, {'pdgid': self.baseid, 'edition': self.edition}):
                    self.cache['summary'].append(PdgSummaryValue(entry._mapping))
        return self.cache['summary']

    def _count_data_entries(self, pdgid, edition):
        """Count number of data entries for a given PDG identifier and edition."""
        pdgdata_table = self.api.db.tables['pdgdata']
        query = select(func.count("*")).select_from(pdgdata_table)
        query = query.where(pdgdata_table.c.pdgid == bindparam('pdgid'))
        query = query.where(pdgdata_table.c.edition == bindparam('edition'))
        with self.api.engine.connect() as conn:
            return conn.execute(query, {'pdgid': pdgid.upper(), 'edition': edition}).scalar()

    def get_parent_pdgid(self, include_edition=True):
        """Return PDG Identifier of this property's parent. In most cases, this
        will be the PDG ID of the particle itself. For those properties, such as
        neutrino mixing angles, that don't have a specific parent particle, the
        parent will be a top-level section (S067 in this case). If this
        property's direct parent is a subsection header, it will be skipped, and
        the PDGID of the top-level section or particle will be returned
        instead."""
        if self._get_pdgid()['parent_pdgid'] is None:
            return None
        p = self
        while p._get_pdgid()['parent_pdgid'] is not None:
            p = self.api.get(p._get_pdgid()['parent_pdgid'], self.edition)
        return p.pdgid if include_edition else p.baseid

    def get_particles(self):
        """Return PdgParticleList for this property's particle."""
        p = self.api.get(self.get_parent_pdgid())
        if p.data_type != 'PART':
            err = 'Identifier %s does not have a parent particle'
            raise PdgNoDataError(err)
        return p

    def get_particle(self):
        """Returns PdgParticle for this property's particle. Raises
        PdgAmbiguousValueError when there are multiple matches."""
        ps = self.get_particles()
        assert len(ps) > 0
        if len(ps) > 1:
            err = "More than one PdgParticle found. Consider using get_particles() instead."
            raise PdgAmbiguousValueError(err)
        return ps[0]

    def get_children(self, recurse=False):
        pdgid_table = self.api.db.tables['pdgid']
        ## NOTE: Querying on IDs doesn't work because the `parent_id` seems off
        # query = select(pdgid_table.c.pdgid) \
        #     .where(pdgid_table.c.parent_id == bindparam('parent_id'))
        # params = {'parent_id': self._get_pdgid()['id']}
        query = select(pdgid_table.c.pdgid) \
            .where(pdgid_table.c.parent_pdgid == bindparam('parent_pdgid'))
        params = {'parent_pdgid': self.baseid}
        with self.api.engine.connect() as conn:
            child_pdgids = [row.pdgid for row
                            in conn.execute(query, params)]
        for child_pdgid in child_pdgids:
            child = self.api.get(child_pdgid)
            yield child
            if recurse:
                for c in child.get_children(recurse=True):
                    yield c

    @property
    def edition(self):
        """Year of edition for which data is requested."""
        return self._edition

    @edition.setter
    def edition(self, edition):
        """Set year of edition used for retrieving data (invalidates cache)."""
        self._edition = edition
        self.pdgid = make_id(self.baseid, self._edition)
        self.cache = dict()

    @property
    def description(self):
        """Description of data."""
        return self._get_pdgid()['description']

    @property
    def data_type(self):
        """Type of data."""
        return self._get_pdgid()['data_type']

    @property
    def data_flags(self):
        """Flags augmenting data type information."""
        return self._get_pdgid()['flags']


class PdgProperty(PdgData):
    """Base class for containers for data containers for particle properties."""

    def summary_values(self, summary_table_only=False):
        """Return list of summary values for this quantity.

        By default, all summary values are included, even if they are only shown in the
        Particle Listings and not listed in the Summary Tables. Set summary_table_only to
        True to get only summary values listed in the Summary Tables.
        """
        if summary_table_only:
            return [v for v in self._get_summary_values() if v.in_summary_table]
        else:
            return self._get_summary_values()

    def n_summary_table_values(self):
        """Return number of summary values in Summary Table for this quantity."""
        return len(self.summary_values(summary_table_only=True))

    def best_summary(self, summary_table_only=False):
        """Return the PDG "best" summary value for this quantity.

        If there is either a single summary value in Particle Listings and Summary Tables, or there are multiple
        summary values but only one is included in the Summary Tables, then this value is returned as the
        PDG best value.

        If there are multiple summary values (e.g. based on assuming or not assuming CPT in the evaluation) and
        PdgApi.pedantic is False, the first value shown in Summary Tables or Particle Listings will be returned.
        If PdgApi.pedantic is True, a PdgAmbiguousValue Exception will be raised.

        If there are no summary values, None is returned.

        If summary_table_only is True, then the best value must be included in the Summary Table and cannot
        be shown only in the Particle Listings.
        """
        if not summary_table_only:
            summaries = self.summary_values(summary_table_only=False)
            if len(summaries) == 1:
                return summaries[0]
            else:
                return self.best_summary(summary_table_only=True)
        else:
            summaries = self.summary_values(summary_table_only=True)
            if len(summaries) == 1:
                return summaries[0]
            elif len(summaries) == 0:
                return None
            else:
                if self.api.pedantic:
                    raise PdgAmbiguousValueError('%s (%s) has multiple summary values' % (self.pdgid, self.description))
                else:
                    return summaries[0]

    def has_best_summary(self, summary_table_only=False):
        """Return True if there is a single PDG "best" value (see best_value() for definition)."""
        try:
            return self.best_summary(summary_table_only) is not None
        except PdgAmbiguousValueError:
            return False

    def get_measurements(self):
        """Return all of the measurements associated with this property."""
        pdgmsmt_table = self.api.db.tables['pdgmeasurement']
        query = select(pdgmsmt_table.c.id)
        query = query.where(pdgmsmt_table.c.pdgid == bindparam('pdgid'))
        with self.api.engine.connect() as conn:
            for entry in conn.execute(query, {'pdgid': self.baseid}):
                yield PdgMeasurement(self.api, entry.id)

    @property
    def confidence_level(self):
        """Shortcut for best_summary().confidence_level."""
        return self.best_summary().confidence_level

    @property
    def is_limit(self):
        """Shortcut for best_summary().is_limit."""
        return self.best_summary().is_limit

    @property
    def value(self):
        """Shortcut for best_summary().value."""
        return self.best_summary().value

    @property
    def error(self):
        """Shortcut for best_summary().error."""
        return self.best_summary().error

    @property
    def error_positive(self):
        """Shortcut for best_summary().error."""
        return self.best_summary().error_positive

    @property
    def error_negative(self):
        """Shortcut for best_summary().error."""
        return self.best_summary().error_negative

    @property
    def scale_factor(self):
        """Shortcut for best_summary().scale_factor."""
        return self.best_summary().scale_factor

    @property
    def units(self):
        """Shortcut for best_summary().units."""
        return self.best_summary().units

    @property
    def comment(self):
        """Shortcut for best_summary().comment."""
        return self.best_summary().comment

    @property
    def value_text(self):
        """Shortcut for best_summary().value_text."""
        return self.best_summary().value_text

    @property
    def display_value_text(self):
        """Shortcut for best_summary().display_value_text."""
        return self.best_summary().display_value_text


class PdgMass(PdgProperty):
    pass


class PdgWidth(PdgProperty):
    pass


class PdgLifetime(PdgProperty):
    pass


class PdgText(PdgData):
    pass



"""
Utilities for PDG API.
"""
import math

from sqlalchemy import select, bindparam

from pdg.errors import PdgNoDataError, PdgAmbiguousValueError, PdgRoundingError


def pdg_round(value, error):
    """Return (value, error) as numbers rounded following PDG rounding rules."""
    # FIXME: might switch to returning decimal.Decimal rather than floats
    if error <= 0.:
        raise PdgRoundingError('PDG rounding requires error larger than zero')
    log = math.log10(abs(error))
    if abs(error) < 1.0 and int(log) != log:
        power = int(log)
    else:
        power = int(log) + 1
    reduced_error = error * 10 ** (-power)
    if reduced_error < 0.355:
        n_digits = 2
    elif reduced_error < 0.950:
        n_digits = 1
    else:
        reduced_error = 0.1
        power += 1
        n_digits = 2
    new_error = round(reduced_error, n_digits) * 10 ** power
    new_value = round(value * 10 ** (-power), n_digits) * 10 ** power
    return new_value, new_error


def parse_id(pdgid):
    """Parse PDG Identifier and return (normalized base identifier, edition)."""
    try:
        baseid, edition = pdgid.split('/')
    except ValueError:
        baseid = pdgid
        edition = None
    return baseid.upper(), edition


def base_id(pdgid):
    """Return normalized base part of PDG Identifier."""
    return parse_id(pdgid)[0]


def make_id(baseid, edition=None):
    """Return normalized full PDG Identifier, possibly including edition."""
    if baseid is None:
        return None
    if edition is None:
        return baseid.upper()
    else:
        return ('%s/%s' % (baseid, edition)).upper()


def best(properties, pedantic=False, quantity=None):
    """Return the "best" property from an iterable of properties, or raise an exception.

    best does (pedantic=False) or does not (pedantic=True) make assumptions about what the best property might
    be in cases where the choice may be ambiguous. In the latter case, a PdgAmbiguous exception is raised while in the
    former case the best property is determined based on data flags and position in the Summary Tables.
    PdgNoDataError will be raised if there is no property that qualifies as best one.

    quantity is an optional string that in case of an exception will describe the property being sought.
    """
    for_what = ' for %s' % quantity if quantity else ''
    props_without_alternates = [p for p in properties if 'A' not in p.data_flags]
    # in non-pedantic mode, filter out "special" values
    if not pedantic:
        props_without_alternates = [p for p in props_without_alternates
                                    if 's' not in p.data_flags]
    if len(props_without_alternates) == 0:
        raise PdgNoDataError('No best property found%s' % for_what)
    elif len(props_without_alternates) == 1:
        return props_without_alternates[0]
    else:
        if pedantic:
            err = 'Ambiguous best property%s' % for_what
            raise PdgAmbiguousValueError(err)
        else:
            props_best = [p for p in props_without_alternates if 'D' in p.data_flags]
            if len(props_best) >= 1:
                return props_best[0]
            else:
                return props_without_alternates[0]


def get_row_data(api, table_name, row_id):
    """Return dict built from the row of the specified table that has an id of
    row_id.
    """
    table = api.db.tables[table_name]
    query = select(table).where(table.c.id == bindparam('id'))
    with api.engine.connect() as conn:
        matches = conn.execute(query, {'id': row_id}).fetchall()
    assert len(matches) == 1
    return matches[0]._mapping


def get_linked_ids(api, table_name, src_col, src_id, dest_col='id'):
    """Return iterator over all values of dest_col in the specified table for which
    src_col = src_id.
    """
    table = api.db.tables[table_name]
    query = select(table.c[dest_col]) \
        .where(table.c[src_col] == bindparam('src_id'))
    with api.engine.connect() as conn:
        for entry in conn.execute(query, {'src_id': src_id}):
            yield entry._mapping[dest_col]



#!/usr/bin/env python

"""
Test cases for measurements.
"""
from __future__ import print_function

import unittest

import pdg


class TestMeasurements(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.api = pdg.connect()

    def test_muon(self):
        muon = self.api.get_particle_by_name('mu-')

        m_msmts = list(muon.mass_measurements(require_summary_data=False))
        t_msmts = list(muon.lifetime_measurements(require_summary_data=False))

        mass = next(m for m in m_msmts
                    if m.comment == '2018 CODATA value'
                    and m.pdgid == 'S004M')
        self.assertEqual(mass.pdgid, 'S004M')
        self.assertEqual(mass.event_count, '')
        self.assertIsNone(mass.confidence_level)
        self.assertEqual(mass.technique, 'RVUE')
        self.assertEqual(mass.charge, '')
        self.assertFalse(mass.changebar)
        self.assertEqual(mass.comment, '2018 CODATA value')

        ref = mass.reference
        self.assertEqual(ref.publication_name, 'RMP 93 025010')
        self.assertEqual(ref.publication_year, 2021)
        self.assertEqual(
            ref.title,
            'CODATA recommended values of the fundamental physical constants: 2018')
        self.assertEqual(ref.doi, '10.1103/RevModPhys.93.025010')
        self.assertEqual(ref.inspire_id, '1898737')

        values = list(mass.values())
        self.assertEqual(len(values), 1)
        value = values[0]
        self.assertEqual(value.measurement.id, mass.id)
        self.assertEqual(value.column_name, 'VALUE')
        # self.assertEqual(value.column_name_tex, 'VALUE')
        self.assertEqual(value.unit_text, 'MeV')
        # self.assertEqual(value.unit_tex, 'MeV')
        self.assertEqual(value.value_text, '105.6583755 +- 0.0000023')
        # self.assertEqual(value.value_tex, '$105.6583755$ $\\pm0.0000023$')
        self.assertEqual(value.display_value_text, '105.6583755 +-0.0000023')
        self.assertEqual(value.display_power_of_ten, 0)
        self.assertFalse(value.display_in_percent)
        self.assertFalse(value.is_limit)
        self.assertFalse(value.is_upper_limit)
        self.assertFalse(value.is_lower_limit)
        self.assertTrue(value.used_in_average)
        self.assertTrue(value.used_in_fit)
        self.assertEqual(round(value.value, 7), 105.6583755)
        self.assertEqual(round(value.error_positive, 7), 0.0000023)
        self.assertEqual(value.error_negative, value.error_positive)
        self.assertIsNone(value.stat_error_positive)
        self.assertIsNone(value.stat_error_negative)
        self.assertIsNone(value.syst_error_positive)
        self.assertIsNone(value.syst_error_negative)

        lifetime = next(t for t in t_msmts if t.comment == 'Surface mu+ at PSI')
        self.assertEqual(lifetime.pdgid, 'S004T')
        self.assertEqual(lifetime.event_count, '')
        self.assertIsNone(lifetime.confidence_level)
        self.assertEqual(lifetime.technique, 'CNTR')
        self.assertEqual(lifetime.charge, '+')
        self.assertFalse(lifetime.changebar)
        self.assertEqual(lifetime.comment, 'Surface mu+ at PSI')
        self.assertEqual(next(lifetime.footnotes()).text,
                         'TISHCHENKO 2013 uses 1.6E12 mu+ events and supersedes WEBBER 2011.')

        ref = lifetime.reference
        self.assertEqual(ref.publication_name, 'PR D87 052003')
        self.assertEqual(ref.publication_year, 2013)
        self.assertEqual(
            ref.title,
            'Detailed Report of the MuLan Measurement of the Positive Muon Lifetime and Determination of the Fermi Constant')
        self.assertEqual(ref.doi, '10.1103/PhysRevD.87.052003')
        self.assertEqual(ref.inspire_id, '1198154')

        values = list(lifetime.values())
        self.assertEqual(len(values), 1)
        value = values[0]
        self.assertEqual(value.measurement.id, lifetime.id)
        self.assertEqual(value.column_name, 'VALUE')
        # self.assertEqual(value.column_name_tex, 'VALUE')
        self.assertEqual(value.unit_text, 's')
        # self.assertEqual(value.unit_tex, 's')
        self.assertEqual(value.value_text, '2.1969803+-0.0000021+-0.0000007 E-6')
        # self.assertEqual(value.value_tex, '($2.1969803$ $\\pm0.0000021$ $\\pm0.0000007$) $ \\times 10^{-6}$')
        self.assertEqual(value.display_value_text, '2.1969803 +-0.0000021 +-0.0000007')
        self.assertEqual(value.display_power_of_ten, -6)
        self.assertFalse(value.display_in_percent)
        self.assertFalse(value.is_limit)
        self.assertFalse(value.is_upper_limit)
        self.assertFalse(value.is_lower_limit)
        self.assertTrue(value.used_in_average)
        self.assertTrue(value.used_in_fit)
        self.assertEqual(round(value.value*1e6, 7), 2.1969803)
        self.assertEqual(round(value.error_positive*1e12, 5), 2.21359)
        self.assertEqual(value.error_negative, value.error_positive)
        self.assertEqual(round(value.stat_error_positive*1e13), 21)
        self.assertEqual(value.stat_error_negative, value.stat_error_positive)
        self.assertEqual(round(value.syst_error_positive*1e13), 7)
        self.assertEqual(value.syst_error_negative, value.syst_error_positive)



"""
Test cases for the info table and other metadata.
"""
from __future__ import print_function

import sys
import os
import unittest
import time
import sqlalchemy

import pdg


class TestMetaData(unittest.TestCase):

    api = None

    @classmethod
    def setUpClass(cls):
        print()
        header = 'Testing with Python %s, SQLAlchemy %s' % (sys.version, sqlalchemy.__version__)
        print(header)
        print('*'*len(header))
        cls.api = pdg.connect()
        if len(cls.api.editions) == 1:
            print('WARNING: testing with single-edition database - not all tests can be run')
        print()
        print(cls.api)

    def test_api_version(self):
        self.assertEqual(pdg.__version__, '0.2.0', 'version number mismatch')

    def test_has_producer(self):
        self.assertIsNotNone(self.api.info('producer'), 'database does not have producer info')

    def test_has_status(self):
        self.assertIsNotNone(self.api.info('status'), 'database does not have status info')

    def test_has_schema_version(self):
        self.assertIsNotNone(self.api.info('schema_version'), 'database does not have schema version info')

    def test_has_created(self):
        self.assertIsNotNone(self.api.info('data_release'), 'database does not have data_release data')

    def test_has_created_timestamp(self):
        self.assertIsNotNone(self.api.info('data_release_timestamp'), 'database does not have data_release timestamp')
        os.environ['TZ'] = 'America/Los_Angeles'
        time.tzset()
        timestamp = time.strftime('%Y-%m-%d %H:%M:%S %Z', time.localtime(float(self.api.info('data_release'))))
        self.assertEqual(self.api.info('data_release_timestamp'), timestamp)

    def test_has_edition(self):
        self.assertIsNotNone(self.api.info('edition'), 'database does not have edition info')

    def test_has_citation(self):
        self.assertIsNotNone(self.api.info('citation'), 'database does not have citation info')

    def test_has_license(self):
        self.assertIsNotNone(self.api.info('license'), 'database does not have license info')

    def test_has_about(self):
        self.assertIsNotNone(self.api.info('about'), 'database does not have about info')

    def test_schema_version(self):
        self.assertTrue(float(self.api.info('schema_version')) >= pdg.MIN_SCHEMA_VERSION)


if __name__ == '__main__':
    unittest.main()



"""
Test cases for utility functions.
"""
from __future__ import print_function

import unittest

from pdg import utils


class TestUtils(unittest.TestCase):

    def test_rounding(self):
        self.assertEqual(utils.pdg_round(0.827, 0.119), (0.83, 0.12))
        self.assertEqual(utils.pdg_round(0.827, 0.367), (0.8, 0.4))
        self.assertEqual(utils.pdg_round(12.3456, .99), (12.3, 1.0))

    def test_parse_id(self):
        self.assertEqual(utils.parse_id('s043m/2020'), ('S043M', '2020'))
        self.assertEqual(utils.parse_id('s043m'), ('S043M', None))

    def test_base_id(self):
        self.assertEqual(utils.base_id('q007/2020'), 'Q007')

    def test_make_id(self):
        self.assertEqual(utils.make_id('q007', None), 'Q007')
        self.assertEqual(utils.make_id('q007', '2020'), 'Q007/2020')


if __name__ == '__main__':
    unittest.main()
